import math
import random
from typing import Union, Tuple

from .entities.pawn import Pawn, _action_decorator
from .entities.characters import Character, Party
from .utilities.location import Point, bresenham, get_adjacent_points

from .armor import ClothArmor, LeatherArmor, PlateArmor, ChainmailArmor, Shield
from .weapons import Sword, Claymore, Mace, Dagger, Staff, ShortBow, Wand, SideKnife, Lute

from .buffs import *
from .debuffs import * 


###############
# ~~ TANKS ~~ #
###############

class Guardian(Character):
    def __init__(self, name: str) -> None:
        '''
        The Guardian class can take a lot of damage, but isn't very good at dealing damage itself.
        '''
        symbol: str = '🛡️'
        role: str='tank'
        position: Union[Point,Tuple[int, int]] = Point(0, 0)
        health_max: int = 110
        gear = PlateArmor()
        super().__init__(name=name, symbol=symbol, role=role, position=position, health_max=health_max, gear=gear)
        self.equip(Sword())
        self.equip(Shield())

    @_action_decorator(melee=True, affected_by_blind=True) # type: ignore
    def defensive_strike(self, target: Pawn) -> None:
        "Strike the target and apply 2 stacks of Toughness to yourself."
        target._take_damage(self, self.calculate_damage(2, target), 'physical')
        self.effects.add_stacks(Toughness, stacks=2, duration=5)

    @_action_decorator(cooldown=10, affected_by_blind=True) # type: ignore
    def shield_stance(self) -> None:
        """
        Take on a defensive stance reducing damage you receive by 75% for 3 turns.
        You are unable to take any action during this time. At the end you
        will receive 5 stacks of Toughness for each hit taken.
        """
        self.effects.add(Stun(3))
        self.effects.add(ShieldStance(self))

    @_action_decorator(cooldown=5) # type: ignore
    def shield_spike(self) -> None:
        """
        Give yourself 50% Reflect for 2 turns. Consumes your Toughness and adds
        it to the Reflect amount.
        """
        extra = self.effects.count('Toughness') * 0.05
        self.effects.add(Reflect(2, extra))
        self.effects.remove_name('Toughness')

    @_action_decorator(cooldown=10, melee=True, affected_by_blind=True) # type: ignore
    def shield_bash(self, target: Pawn) -> None:
        """
        Bash the target with your shield and apply Stun for 1 turn.
        """
        target._take_damage(self, self.calculate_damage(10, target), 'physical')
        target.effects.add(Stun(1))

    @_action_decorator(cooldown=100) # type: ignore
    def inspiration(self, party: Party) -> None:
        """
        All damage the Guardian receives will instead heal the Party and apply
        1 stack of Toughness for 5 turns.
        """
        self.effects.add(Inspiration(party))


class Paladin(Character):
    def __init__(self, name: str) -> None:
        '''
        Powerful holy warrior that strikes a balance between offense and defense.
        
        '''
        symbol: str = '⚜️'
        role: str='tank'
        position: Union[Point,Tuple[int,int]] = Point(0, 0)
        health_max: int = 120
        gear = PlateArmor()
        super().__init__(name=name, symbol=symbol, role=role, position=position, health_max=health_max, gear=gear)
        self.equip(Sword())
        self.equip(Shield())
    
    @_action_decorator(melee=True, affected_by_blind=True) # type: ignore
    def holy_strike(self, target: Pawn) -> None:
        'Strike the target and apply 1 stack of Frailty.'
        target._take_damage(self, self.calculate_damage(5, target), 'physical')
        target.effects.add(Frailty(5))

    @_action_decorator(cooldown=5, affected_by_blind=False) # type: ignore
    def exorcise(self, target: Pawn) -> None:
        '''
        consume all stacks of Frailty to give yourself double the
        stacks in Toughness for 3 turns
        '''
        target._take_damage(self, self.calculate_damage(10, target), 'spirit')
        stacks = target.effects.count('Frailty')
        self.effects.add_stacks(Toughness, stacks=stacks*2, duration=3)
        target.effects.remove_name('Frailty')

    @_action_decorator(cooldown=10, affected_by_blind=False) # type: ignore
    def divine_protection(self, target: Pawn) -> None:
        '''
        Apply a Barrier to target or self to absorb the next hit and store the damage.
        '''
        target.effects.add(DivineProtection(caster=self, duration=10))

    @_action_decorator(cooldown=10, affected_by_blind=False) # type: ignore
    def prayer(self, party: Party) -> None:
        '''
        Remove all stacks of Toughness and buff the Party increasing the power of their next attack by the amount of Toughness consumed.
        '''
        stacks = self.effects.count('Toughness')
        self.effects.remove_name('Toughness')
        for member in party:
            member.effects.add(NextAttack(0.05 * stacks))

    @_action_decorator(cooldown=100, affected_by_blind=False) # type: ignore
    def holy_nova(self, target: Pawn, party: Party) -> None:
        '''
        Damage absorbed by Holy Barrier is used to trigger a holy nova on your target dealing damage and healing all Party members. Applies your Toughness stacks to the group for 5 turns.
        '''
        barrier = self.calculate_damage(self.reports.get('barrier', 0), target)
        self.reports['barrier'] = 0
        stacks = self.effects.count('Toughness')
        self.effects.remove_name('Toughness')
        for member in party:
            member.effects.add_stacks(Toughness, stacks=stacks, duration=5)
            member._heal(barrier)

        target._take_damage(self, barrier, 'spirit')

    
class Berzerker(Character):
    def __init__(self, name:str) -> None:
        symbol: str = '⚔️'
        role: str='tank'
        position: Union[Point,Tuple[int,int]] = Point(0, 0)
        health_max: int = 150
        gear = ChainmailArmor()
        super().__init__(name=name, symbol=symbol, role=role, position=position, health_max=health_max, gear=gear)
        self.equip(Claymore())

    @_action_decorator(cooldown=10, affected_by_root=True, affected_by_blind=True) # type: ignore
    def charge(self, target: Pawn) -> None:
        'Charge to the front of the target in a single turn.'
        self._teleport(target.facing_direction)
        self.face(target.position)

    @_action_decorator(melee=True, affected_by_blind=True) # type: ignore
    def frenzied_attack(self, target: Pawn) -> None:
        'Attack the target with no regard for your own safety. Apply 1 stack of Might and Expose Weakness to self.'
        target._take_damage(self, self.calculate_damage(15, target), 'physical')
        self.effects.add(Might(float('inf')))
        self.effects.add(ExposeWeakness(float('inf')))
    
    @_action_decorator(cooldown=10, melee=True, affected_by_blind=True) # type: ignore
    def rampage(self, target: Pawn) -> None:
        '''
        A devastating attack that applies 4 stacks of Might and Expose Weakness to yourself but cripples the enemy and applies 4 stacks of Magic Vulnerability, Expose Weakness,  and also Poison Vulnerability.
        '''
        self.effects.add_stacks(Might, stacks=4, duration=float('inf'))
        self.effects.add_stacks(ExposeWeakness, stacks=4, duration=float('inf'))

        target._take_damage(self, self.calculate_damage(30, target), 'physical')
        target.effects.add_stacks(PoisonVulnerability, stacks=4, duration=2)
        target.effects.add_stacks(ExposeWeakness, stacks=4, duration=2)
        target.effects.add_stacks(MagicVulnerability, stacks=4)

    @_action_decorator(cooldown=10, affected_by_blind=True) # type: ignore
    def parry(self) -> None:
        self.effects.add(Parry(user=self, duration=1))

    @_action_decorator(cooldown=100, affected_by_blind=False) # type: ignore
    def battle_shout(self, party: Party) -> None:
        stacks = self.effects.count('Might')
        self.effects.remove_name('Might')
        for member in party:
            member.effects.add_stacks(Might, stacks=stacks, duration=10)

    
#################
# ~~ HEALERS ~~ #
#################

class Cleric(Character):
    '''
    A dedicated healer that can heal and cleanse the party of debuffs. 
    '''

    def __init__(self, name: str) -> None:
        symbol: str = '🔅'
        role: str='healer'
        position: Union[Point,Tuple[int, int]] = Point(0, 0)
        health_max: int = 100
        gear = PlateArmor()
        super().__init__(name=name, symbol=symbol, role=role, position=position, health_max=health_max, gear=gear)
        self.equip(Mace())
        self.equip(Shield())

    @_action_decorator
    def healing_word(self, target: Pawn) -> None:
        x = self.calculate_damage(target.health_max//2, target)
        target._heal(x)

    @_action_decorator(cooldown=10, affected_by_blind=True) # type: ignore
    def jolt(self, target: Pawn) -> None:
        if target.has_effect(MagicVulnerability()):
            target.effects.add(Stun(3))
            target.effects.remove_all(MagicVulnerability())
        else:
            target.effects.add(Stun(2))

    @_action_decorator(affected_by_blind=True) # type: ignore
    def smite(self, target: Pawn) -> None:
        if self.reports.get('barrier'):
            target.effects.add_stacks(Frailty, stacks=5, duration=3)
            target._take_damage(self,
                self.calculate_damage(10, target) + self.reports['barrier'],
                'spirit')
            self.reports['barrier'] = 0
        else:
            target.effects.add(Frailty(3))
            target._take_damage(self, self.calculate_damage(5, target), 'spirit')

    @_action_decorator # type: ignore
    def cleanse(self, target: Pawn) -> None:
        target.effects.remove_category('curable')

    @_action_decorator(cooldown=100) # type: ignore
    def group_barrier(self, party: Party) -> None:
        for member in party.members:
            member.effects.add(Barrier(self, 4))


class Druid(Character):
    def __init__(self, name:str) -> None:
        symbol: str = '☘️'
        role: str='healer'
        position: Union[Point,Tuple[int,int]] = Point(0, 0)
        health_max: int = 100
        gear = LeatherArmor()
        super().__init__(name=name, symbol=symbol, role=role, position=position, health_max=health_max, gear=gear)
        self.equip(Staff())

    @_action_decorator(affected_by_blind=False) # type: ignore
    def swift_mend(self, target: Pawn) -> None:
        heal_amount = self.calculate_damage(target.health_max//3, target)
        target._heal(heal_amount)
        target.effects.add(HoT(name="Swift Mend", duration=2, heal_amount=heal_amount//2))
    
    @_action_decorator(cooldown=10, affected_by_blind=True) # type: ignore
    def grasping_roots(self, target: Pawn) -> None:
        target.effects.add(Root(4))
        if target.has_effect(MagicVulnerability()):
            target.effects.add_stacks(ExposeWeakness, stacks=5, duration=3)
            target.effects.remove_all(MagicVulnerability())

    @_action_decorator(affected_by_blind=True) # type: ignore
    def blaze(self, target: Pawn) -> None:
        '''
        Blasts the target with intense heat and apply 1
        stack of Expose Weakness and Fire Resistance
        '''
        target._take_damage(self, self.calculate_damage(10, target), 'fire')
        target.effects.add(FireResistance())
        target.effects.add(ExposeWeakness(3))

    @_action_decorator(affected_by_blind=True) # type: ignore
    def salve(self, target: Pawn) -> None:
        'Remove all curable effects from target'
        target.effects.remove_category('curable')

    @_action_decorator(cooldown=100) # type: ignore
    def group_thornshield(self, party: Party) -> None:
        'Apply Reflect and Toughness to party for 8 turns'
        for member in party.members:
            member.effects.add(Reflect(8))
            member.effects.add_stacks(Toughness, stacks=5, duration=8)


class Shaman(Character):
    def __init__(self, name:str) -> None:
        symbol: str = '🍄'
        role: str='healer'
        position: Union[Point,Tuple[int,int]] = Point(0, 0)
        health_max: int = 100
        gear = ChainmailArmor()
        super().__init__(name=name, symbol=symbol, role=role, position=position, health_max=health_max, gear=gear)
        self.equip(Staff())

    @_action_decorator(affected_by_blind=False) # type: ignore
    def regenerate(self, target: Pawn) -> None:
        'Heal target for 80% of their health max over 5 turns'
        heal_amount = self.calculate_damage(target.health_max // 5 * 4, target)
        target.effects.add(HoT(name="Regenerate", duration=5, heal_amount=heal_amount//5))

    @_action_decorator(cooldown=5, affected_by_blind=True) # type: ignore
    def enhancement(self, target: Pawn) -> None:
        'Apply 5 stacks of Might and Toughness to the target for 3 turns'
        target.effects.add_stacks(Might, stacks=5, duration=3)
        target.effects.add_stacks(Toughness, stacks=5, duration=3)

    @_action_decorator(cooldown=2, affected_by_blind=True) # type: ignore
    def poison_frost(self, target: Pawn) -> None:
        '''
        Apply a frost based poison to the target, 2 stacks of
        Frailty, and 1 stack of Frost Resistance
        '''
        target.effects.add(Poison(target, 4, self.calculate_damage(4, target)))
        target.effects.add_stacks(Frailty, stacks=2, duration=4)
        target.effects.add(FrostResistance())
        if target.vulnerable_to('poison'):
            target.effects.add(Poison(target, 4, self.calculate_damage(4, target)))
            target.effects.add_stacks(Frailty, stacks=2, duration=4)
            target.effects.add(FrostResistance())
    
    @_action_decorator(affected_by_blind=False) # type: ignore
    def purge(self, target: Pawn) -> None:
        'Remove all curable effects from target'
        target.effects.remove_category('curable')

    @_action_decorator(cooldown=100) # type: ignore
    def group_infusion(self, party: Party) -> None:
        'Apply 10 stacks of Might to Party for 8 turns'
        for member in party.members:
            member.effects.add_stacks(Might, stacks=10, duration=8)

    

#############
# ~~ DPS ~~ #
#############

class Ranger(Character):
    def __init__(self, name:str):
        symbol: str = '🏹'
        role: str='dps'
        position: Union[Point,Tuple[int,int]] = Point(0, 0)
        health_max: int = 80
        gear = LeatherArmor()
        super().__init__(name=name, symbol=symbol, role=role, position=position, health_max=health_max, gear=gear)
        self.equip(ShortBow())
        self.reports['shots'] = 0

    @_action_decorator(melee=False, affected_by_blind=True) #type: ignore
    def shoot(self, target: Pawn) -> None:
        dmg = self.calculate_damage(len(list(bresenham(self.position, target.position))), target)
        self.reports['shots'] += 1
        if self.reports['shots'] >= 3:
            target.effects.add(ExposeWeakness(2))
        target._take_damage(self, dmg, "physical")

    @_action_decorator(cooldown=5, melee=False, affected_by_blind=True) #type: ignore
    def virulent_arrow(self, target: Pawn) -> None:
        self.reports['shots'] += 1
        if not target.poisoned:
            target.effects.add(PoisonVulnerability(3))
        target.effects.add(Poison(target, duration=3, dot_amount=5))
        target._take_damage(self, self.calculate_damage(10, target), "physical")

    @_action_decorator(cooldown=10, melee=False, affected_by_blind=True) #type: ignore
    def frostfire_arrow(self, target: Pawn) -> None:
        self.reports['shots'] += 1
        if not target.has_effect(FireResistance()) and not target.has_effect(FrostResistance()):
            target.effects.add_stacks(MagicVulnerability, stacks=3)
        target._take_damage(self, self.calculate_damage(10, target), "physical")

    @_action_decorator(cooldown=5, melee=False) #type: ignore
    def field_medicine(self, target: Pawn) -> None:
        self.reports['shots'] = 0
        target._heal(self.calculate_damage(target.health_max//5, target))
        target.effects.remove_category('curable')

    @_action_decorator(cooldown=100, melee=False) #type: ignore
    def murder(self, target: Pawn) -> None:
        'Send a murder of crows to enact judgement on a target.'
        self.reports['shots'] = 0
        target.effects.add(Blind(4))
        target.effects.add_stacks(ExposeWeakness, stacks=6, duration=4)
        target.effects.add(DoT(target, name="Murder of Crows", duration=4, dot_amount=self.calculate_damage(10, target)))


class Rogue(Character):
    def __init__(self, name) -> None:
        symbol: str = '🗡️'
        role: str='dps'
        position: Union[Point,Tuple[int,int]] = Point(0, 0)
        health_max: int = 100
        gear = LeatherArmor()
        super().__init__(name=name, symbol=symbol, role=role, position=position, health_max=health_max, gear=gear)
        self.equip(Dagger())
        self.equip(SideKnife())
        self.reports['backstab'] = 0

    def is_behind(self, target: Pawn) -> bool:
        return self.position in target.points_behind

    @_action_decorator(melee=True, affected_by_blind=True) #type: ignore
    def backstab(self, target: Pawn) -> None:
        "Try doing it from behind!"
        dmg = self.calculate_damage(20, target)
        if self.is_behind(target):
            dmg *= 2
            self.reports['backstab'] += 1
            if self.reports['backstab'] >= 3:
                target.effects.add_stacks(ExposeWeakness, stacks=2, duration=3)
        target._take_damage(self, dmg, "physical")

    @_action_decorator(cooldown=5, melee=True) #type: ignore
    def envenom(self, target: Pawn) -> None:
        "Poisons and Enfeebles the target"
        self.reports['backstab'] = 0
        target.effects.add(Poison(target, duration=4, dot_amount=5))
        target.effects.add_stacks(Frailty, stacks=4, duration=4)
        if target.effects.vulnerable_to('poison'):
            target.effects.add_stacks(Frailty, stacks=4, duration=4)

    @_action_decorator(cooldown=10, melee=True) #type: ignore
    def sand(self, target: Pawn) -> None:
        "Go for the eyes!"
        self.reports['backstab'] = 0
        target.effects.add(Blind(1))
    
    @_action_decorator(cooldown=10, melee=True) #type: ignore
    def shank(self, target: Pawn) -> None:
        "Twist the knife!"
        self.reports['backstab'] = 0
        target.effects.add_stacks(ExposeWeakness, stacks=target.effects.count('expose_weakness'), duration=2)
        target.effects.add_stacks(MagicVulnerability, stacks=target.effects.count('magic vulnerability'))
        target._take_damage(self, self.calculate_damage(10, target), "physical")

    @_action_decorator(cooldown=100, melee=True) #type: ignore
    def ambush(self, target: Pawn) -> None:
        "Surprise!"
        target.effects.add(Blind(2))
        target.effects.add_stacks(Frailty, stacks=4, duration=10)
        target.effects.add(PoisonVulnerability(5))

        dmg = self.calculate_damage(40, target)
        if self.is_behind(target):
            dmg *= 10
            self.reports['backstab'] += 1
            if self.reports['backstab'] >= 3:
                target.effects.add_stacks(ExposeWeakness, stacks=2, duration=3)
        target._take_damage(self, dmg, "physical")


class Necromancer(Character):
    def __init__(self, name: str) -> None:
        symbol: str = '🦴'
        role: str='dps'
        position: Union[Point,Tuple[int,int]] = Point(0, 0)
        health_max: int = 80
        gear = ClothArmor()
        super().__init__(name=name, symbol=symbol, role=role, position=position, health_max=health_max, gear=gear)
        self.equip(Wand())
    
    @_action_decorator(melee=False) #type: ignore
    def poison_blast(self, target: Pawn) -> None:
        '''
        Blast the target with Poison and apply 1 stack of Expose Weakness
        '''
        target._take_damage(self, self.calculate_damage(20, target), "poison")
        target.effects.add_stacks(ExposeWeakness, stacks=1, duration=4)

    @_action_decorator(cooldown=10, melee=False) #type: ignore
    def doom(self, target: Pawn) -> None:
        '''
        Consume and combine the negative effects on the target into a delayed damage effect that triggers 15 turns. 
        '''
        doom = target.effects.find_effect_exact_text('doom')
        if doom:
            doom.increase_doom() # type: ignore
        else:
            target.effects.add(Doom(self, target, self.calculate_damage(10, target)))

    @_action_decorator(melee=False) #type: ignore
    def blood_magic(self, target: Pawn, sacrifice: int) -> None:
        '''
        Sacrifice health to damagea target. 
        '''
        if sacrifice > self.health:
            sacrifice = self.health
        target._take_damage(self, self.calculate_damage(sacrifice, target), "spirit")
        self.health -= sacrifice

    @_action_decorator(cooldown=10, melee=False) #type: ignore
    def life_tap(self, target: Pawn) -> None:
        '''
        Leach health from the target and transfer it to yourself and apply 2 stacks of Expose Weakness. If the target has Magic Vulnerability apply Poison Vulnerability for 5 turns.
        '''
        dmg = self.calculate_damage(20, target)
        target._take_damage(self, dmg, "spirit")
        self._heal(dmg)
        target.effects.add_stacks(ExposeWeakness, stacks=2, duration=4)
        if target.effects.vulnerable_to('magic'):
            target.effects.add(PoisonVulnerability(5))
    
    @_action_decorator(cooldown=100, melee=False) #type: ignore
    def blood_ritual(self, party: Party, target: Pawn) -> None:
        '''
        Sacrifice 50% health from your entire Party to summon a shade of death that moves towards the target. Once reached it will infect the targets soul with a crippling DoT that also applies 10 stacks of Expose Weakness, Magic Vulnerability, and Frailty for 5 turns.
        '''
        dmg = 0
        for member in party.members:
            dmg += member.health // 2
            member.health -= member.health // 2

        target.effects.add(ShadeOfDeath(self, target, self.calculate_damage(dmg, target)))


class Wizard(Character):
    def __init__(self, name:str) -> None:
        symbol: str = '🧙'
        role: str='dps'
        position: Union[Point,Tuple[int,int]] = Point(0, 0)
        health_max: int = 80
        gear = ClothArmor()
        super().__init__(name=name, symbol=symbol, role=role, position=position, health_max=health_max, gear=gear)
        self.equip(Wand())

    @_action_decorator(melee=False) #type: ignore
    def magic_missile(self, target: Pawn) -> None:
        '''
        Blast the target with magic missiles! If the target has Magic Vulnerability
        this does 50% additional damage and applies Stun for 1 turn.
        '''
        dmg = self.calculate_damage(20, target)
        if target.effects.vulnerable_to('magic'):
            dmg = self.calculate_damage(100, target)
            target.effects.add(Stun(1))
        target._take_damage(self, dmg, "magic")

    @_action_decorator(cooldown=10, melee=False) #type: ignore
    def fire_bolt(self, target: Pawn) -> None:
        'Blast the target with a ball of Fire and apply 2 stacks of Fire Resistance.'
        target._take_damage(self, self.calculate_damage(50, target), "fire")
        target.effects.add_stacks(FireResistance, stacks=5)

    @_action_decorator(cooldown=10, melee=False) #type: ignore
    def frost_bolt(self, target: Pawn) -> None:
        'Blast the target with a ball of Frost and apply 2 stacks of Frost Resistance.'
        target._take_damage(self, self.calculate_damage(50, target), "frost")
        target.effects.add_stacks(FrostResistance, stacks=5)

    @_action_decorator(cooldown=10, melee=False) #type: ignore
    def teleport(self, target: Pawn, location: Union[Point,Tuple[int,int]]) -> None:
        'Instantly Teleport self or target to anywhere in the arena.'
        target._teleport(location)
        target.face(random.choice(get_adjacent_points(location)))

    @_action_decorator(cooldown=100, melee=False) #type: ignore
    def lightning_strike(self, target: Pawn) -> None:
        '''
        Call down a powerful lightning strike on the target! If the target
        has Magic Vulnerability this deals double damage, applies Stun, and
        5 stacks of Magic Vulnerability on the target for 3 turns.
        '''
        dmg = self.calculate_damage(100, target)
        if target.effects.vulnerable_to('magic'):
            dmg *= 10
            target.effects.add(Stun(3))
            target._take_damage(self, dmg, "magic")
            target.effects.add_stacks(MagicVulnerability, stacks=5)
        else:
            target._take_damage(self, dmg, "magic")


class Bard(Character):
    def __init__(self, name: str) -> None:
        symbol: str = '🎵'
        role: str='dps'
        position: Union[Point,Tuple[int,int]] = Point(0, 0)
        health_max: int = 90
        gear = LeatherArmor()
        super().__init__(name=name, symbol=symbol, role=role, position=position, health_max=health_max, gear=gear)
        self.equip(Lute())

    @_action_decorator(melee=False, affected_by_blind=False) #type: ignore
    def healing_notes(self, party: Party) -> None:
        'All Party members heal for 10% of their health'
        for member in party.members:
            member._heal(member.health//10)

    @_action_decorator(melee=False, affected_by_blind=False) #type: ignore
    def violent_notes(self, party: Party) -> None:
        'All Party members receive 5 stacks of Might'
        for member in party.members:
            member.effects.add_stacks(Might, stacks=5, duration=1)

    @_action_decorator(melee=False, affected_by_blind=False) #type: ignore
    def curative_notes(self, party: Party) -> None:
        'All Party members will receive Cure in 2 turns'
        for member in party.members:
            member.effects.add(CurativeNotes(member))

    @_action_decorator(cooldown=5, melee=False, affected_by_blind=False) #type: ignore
    def magical_notes(self, target: Pawn) -> None:
        '''
        Play a piercing magical note at the enemy that deals Magic damage
        and applies 4 stacks of Expose Weakness. Increased to 6 stacks and
        applies Stun if the target has Magic Vulnerability.
        '''
        dmg = self.calculate_damage(20, target)
        if target.effects.vulnerable_to('magic'):
            target.effects.add(Stun(1))
            target.effects.add_stacks(ExposeWeakness, stacks=6, duration=5)
        else:
            target.effects.add_stacks(ExposeWeakness, stacks=4, duration=5)
        target._take_damage(self, dmg, "magic")

    @_action_decorator(cooldown=100, melee=False, affected_by_blind=False) #type: ignore
    def crescendo(self, party: Party, enemy: Pawn) -> None:
        '''
        Combine all your notes into a chaotic lullaby that triggers all bard
        songs and the Next Attack for all group members will do double damage.
        '''
        # All bard songs + Next Attack buff
        for member in party.members:
            member._heal(member.health//10)
            member.effects.add_stacks(Might, stacks=5, duration=1)
            member.effects.add(CurativeNotes(member))
            member.effects.add(NextAttack(1.0))

        dmg = self.calculate_damage(20, enemy)
        if enemy.effects.vulnerable_to('magic'):
            enemy.effects.add(Stun(1))
            enemy.effects.add_stacks(ExposeWeakness, stacks=6, duration=5)
        else:
            enemy.effects.add_stacks(ExposeWeakness, stacks=4, duration=5)
        enemy._take_damage(self, dmg, "magic")


import math
from collections import Counter
from dataclasses import dataclass, field
from functools import singledispatchmethod, wraps
from typing import Any, Literal, Tuple, Union

from ..armor import ClothArmor
from ..debuffs import MagicVulnerability

from .equipment import Gear, GearSet
from .equipment import Equipment
from .effects import Effect, Effects

from ..utilities.location import Point, bresenham, clean_name, distance_between, behinds


@dataclass
class _Character:
    position: Point


##############################
# ~~~ Ability Decorators ~~~ #
##############################

def _check_can_move(func):
    @wraps(func)
    def wrapper(self: 'Pawn', *args, **kwargs):
        
        if not self.rooted and not self.stunned:
            return func(self, *args, **kwargs)

        reason = ''
        if self.effects.rooted:
            reason = f"{self.name} is rooted!"
        if self.effects.stunned:
            reason = f"{self.name} is stunned!"
        self.action_history.append(
            Action(
                turn=self._turn,
                type='move',
                action_name=func.__name__,
                actor=self,
                target=kwargs.get('target', args[0] if args else None),
                failed=True,
                failed_reason=reason
            )
        )

    return wrapper


def _action_decorator(_func=None, *, cooldown: int = 1, melee: bool = False, affected_by_blind: bool = True, affected_by_root: bool = False):
    def actual_decorator(func):
        @wraps(func)
        def wrapper(self: 'Pawn', *args, **kwargs):
            target: Pawn = kwargs.get('target', args[0] if args else self)

            reason = ''
            if self._is_dead:
                reason = f"{self.name} is dead!"
            elif self.effects.stunned:
                reason = f"{self.name} is stunned!"
            elif self.acted_this_turn:
                reason = f"{self.name} has already acted this turn!"
            elif self._ability_cooldowns.get(clean_name(func.__name__), 0) > 0:
                reason = f"{clean_name(func.__name__)} is on cooldown!"
            elif affected_by_root and self.effects.rooted:
                reason = f"{self.name} is rooted!"
            elif affected_by_blind and self.effects.blinded:
                reason = f"{self.name} is blinded!"
            elif melee and distance_between(self.position, target.position) > 1.5:
                reason = f"{target.name} was too far away"

            if reason:
                self.action_history.append(
                    Action(
                        turn=self._turn,
                        type='ability',
                        action_name=func.__name__,
                        actor=self,
                        target=target,
                        failed=True,
                        failed_reason=reason
                    )
                )
                return

            if self._ability_cooldowns.get(clean_name(func.__name__), 0) == 0:
                self._ability_cooldowns[clean_name(func.__name__)] = cooldown
            self.acted_this_turn = True

            self.current_action = Action(
                turn=self._turn,
                type='ability',
                action_name=func.__name__,
                actor=self,
                target=target
            )
            self.action_history.append(self.current_action)
            if target != self:
                self.face(target if not hasattr(target, 'position') else target.position)
            return func(self, *args, **kwargs)

        return wrapper

    if _func is None:
        return actual_decorator
    return actual_decorator(_func)


class Pawn(_Character):

    def __init__(self,
                 name,
                 position: Union[Point, Tuple[int, int]],
                 health_max: int,
                 symbol: str = '@',
                 gear: GearSet = ClothArmor()) -> None:

        self.name = name

        # position
        self._position = position if isinstance(
            position, Point) else Point(*position)
        self.facing_direction = Point(0, 0)

        # status
        self.health_max = health_max
        self._health = health_max
        self._is_dead: bool = False
        self._is_already_dead: bool = False
        self._was_hit: bool = False
        self.current_action: Union[Action, None] = None
        self._ability_cooldowns = {}

        # effects
        self.equipment = Equipment()
        self.equipment.equip(gear)
        self.effects = Effects()

        # internal use
        self.move_history = [self.position]
        self.action_history: list[Action] = []
        self._symbol = symbol
        self._turn = 0
        self.acted_this_turn = False
        self.moved_this_turn = False
        self.reports: dict[str, Any] = {}

    ####################
    # ~~~ Location ~~~ #
    ####################
    @property
    def position(self) -> Point:
        return self._position

    @position.setter
    def position(self, value: Union[Point, Tuple[int, int]]) -> None:
        if not self.moved_this_turn:
            self._position = value if isinstance(
                value, Point) else Point(*value)
            self.acted_this_turn = True
            self.moved_this_turn = True
            self.move_history.append(self.position)
            self.action_history.append(
                Action(
                    turn=self._turn,
                    type='move',
                    action_name='move',
                    actor=self,
                    target=self.position
                )
            )

    def _revert_position(self, message: str) -> None:
        bad_position = self.position
        self.move_history = self.move_history[:-1]
        self._position = self.move_history[-1]
        self.action_history.append(
            Action(
                turn=self._turn,
                type="move",
                action_name="move",
                actor=self,
                target=bad_position,
                failed=True,
                failed_reason=message
            )
        )

    @property
    def points_behind(self) -> Tuple[Point, ...]:
        return behinds(self.position, self.facing_direction)

    @property
    def positions_behind(self) -> Tuple[Point, ...]:
        return self.points_behind

    @property
    def symbol(self) -> str:
        if self.is_alive:
            return self._symbol
        return "💀"

    ##################
    # ~~~ Status ~~~ #
    ##################

    @property
    def cooldowns(self) -> dict[str,int]:
        return {name : cooldown
                for name, cooldown in self._ability_cooldowns.items()
                if cooldown > 0}

    def is_on_cooldown(self, ability_name: str) -> bool:
        '''
        Checks the cooldowns for the name of an ability.
        Returns True if the ability is on cooldown.
        '''
        return self._ability_cooldowns.get(clean_name(ability_name), 0) > 0

    def get_cooldown(self, ability_name: str) -> int:
        '''
        Get the cooldown for an ability.
        Returns 0 if the ability is not on cooldown.
        '''
        return self._ability_cooldowns.get(clean_name(ability_name), 0)

    @property
    def health(self):
        return self._health

    @health.setter
    def health(self, value: int) -> None:
        if not self._is_dead:
            self._health = value
            if self._health < 0:
                self._health = 0
            self._is_dead = self._health <= 0

    @property
    def health_percent(self) -> float:
        return round(self.health / self.health_max, 2)

    @property
    def is_alive(self) -> bool:
        return not self._is_dead

    @property
    def last_action_failed(self) -> bool:
        return bool(self.action_history) and self.action_history[-1].failed

    ########################
    # ~~~ Measurements ~~~ #
    ########################

    def distance_to(self, other: Union['Pawn', _Character, Point, tuple]) -> float:
        '''
        Get the distance between this character and another character or a board square.
        '''
        if hasattr(other, 'position'):
            return distance_between(self.position, other.position) # type: ignore
        elif isinstance(other, Point):
            return distance_between(self.position, other)
        elif isinstance(other, tuple):
            return distance_between(self.position, Point(*other))
        raise TypeError(f"Cannot calculate distance to {other} of type {type(other)}")

    def distance_from(self, other: Union['Pawn', _Character, Point, tuple]) -> float:
        return self.distance_to(other) # type: ignore

    def is_behind(self, other: 'Pawn') -> bool:
        return other.position in self.points_behind

    ####################
    # ~~~ Movement ~~~ #
    ####################
    def move(self, destination: Point) -> None:
        if not self._is_dead:
            return self.move_toward(destination)

    @_check_can_move
    def move_toward(self, target: Union[Point, _Character, tuple]) -> None:
        """
        Move one square in a straight line toward the provided Pawn or Point (x,y) tuple,
        as calculated using Bresenham's algorithm. Does not take into account impassible
        or dangerous terrain -- that is, using this movement method, the pawn might
        attempt to move into a wall and fail the move. Example:
        `hero.move_toward( boss )` or `hero.move_toward( Point(5, 5) )`"""
        if not self._is_dead:
            if isinstance(target, _Character):
                target = target.position
            elif isinstance(target, tuple):
                target = Point(*target)

            if self.distance_from(target) > 1.5:
                path = bresenham(self.position, target)
                self.position = next(path)
                self.face(next(path))
            else:
                add = (target.x-self.position.x, target.y-self.position.y)
                self.position = target
                self.face(Point(
                    target.x+(add[0]),
                    target.y+(add[1])
                ))

    def _teleport(self, point: Union[Point,tuple[int,int]]) -> None:
        if hasattr(point, 'position'):
            point = point.position # type: ignore
        elif isinstance(point, tuple):
            point = Point(*point)

        self.move_history.append(point) # type: ignore
        if self.distance_from(point) > 1.5:
            self.position = point
            return
        self.facing_direction = Point(
            point.x+(point.x-self.position.x), point.y+(point.y-self.position.y)) # type: ignore
        self.position = point

    def face(self, target: Union[_Character,Point,tuple[int,int]]) -> None:
        if hasattr(target, 'position'):
            self.facing_direction = next(bresenham(self.position, target.position)) # type: ignore
        elif isinstance(target, tuple):
            self.facing_direction = next(bresenham(self.position, Point(*target)))
        else:
            self.facing_direction = next(bresenham(self.position, target)) # type: ignore

    @_check_can_move
    def move_up(self) -> None:
        if not self._is_dead:
            self.position = Point(self.position.x, self.position.y+1)
            self.face((self.position.x, self.position.y+1))

    @_check_can_move
    def move_left(self) -> None:
        if not self._is_dead:
            self.position = Point(self.position.x-1, self.position.y)
            self.face((self.position.x-1, self.position.y))

    @_check_can_move
    def move_right(self) -> None:
        if not self._is_dead:
            self.position = Point(self.position.x+1, self.position.y)
            self.face((self.position.x+1, self.position.y))

    @_check_can_move
    def move_down(self) -> None:
        if not self._is_dead:
            self.position = Point(self.position.x, self.position.y-1)
            self.face((self.position.x, self.position.y-1))

    @_check_can_move
    def move_down_right(self) -> None:
        if not self._is_dead:
            self.position = Point(self.position.x+1, self.position.y-1)
            self.face((self.position.x+1, self.position.y-1))

    @_check_can_move
    def move_up_right(self) -> None:
        if not self._is_dead:
            self.position = Point(self.position.x+1, self.position.y+1)
            self.face((self.position.x+1, self.position.y+1))

    @_check_can_move
    def move_down_left(self) -> None:
        if not self._is_dead:
            self.position = Point(self.position.x-1, self.position.y-1)
            self.face((self.position.x-1, self.position.y-1))

    @_check_can_move
    def move_up_left(self) -> None:
        if not self._is_dead:
            self.position = Point(self.position.x-1, self.position.y+1)
            self.face((self.position.x-1, self.position.y+1))

    ##################
    # ~~~ Combat ~~~ #
    ##################

    @property
    def _base_damage(self) -> int:
        return self.equipment.bonus_damage_output + int(round(self.equipment.bonus_damage_output * self.equipment.bonus_damage_output_percent)) + sum([e.deal_bonus_damage_amount for e in self.effects.deal_damage_effects])

    @property
    def _damage_multiplier(self) -> float:
        return sum([e.deal_bonus_damage_percent for e in self.effects.deal_damage_effects])

    def calculate_damage(self, damage, target) -> int:
        dmg = damage + self._base_damage
        for effect in self.effects.deal_damage_effects:
            effect.on_activate(user=self, total_damage=dmg, target=target)
        return dmg + math.ceil(dmg * self._damage_multiplier)

    def _tick_damage(self, effect: Effect) -> None:
        if effect.damage_over_time > 0:
            self.health -= effect.damage_over_time
            self.action_history.append(Action(
                turn=self._turn,
                type='damage',
                action_name=f'{effect.name} ticked for {effect.damage_over_time} damage',
                actor=effect.name,
                target=self,
                target_effects=self.effects.to_dict(),
                ability_used=effect.name,
                damage=effect.damage_over_time,
            ))
        
        if effect.heal_over_time:
            self._heal(effect.heal_over_time)
            self.action_history.append(Action(
                turn=self._turn,
                type='damage',
                action_name=f'{effect.name} healed for {effect.heal_over_time} health',
                actor=effect.name,
                target=self,
                target_effects=self.effects.to_dict(),
                ability_used=effect.name,
                damage=effect.heal_over_time,
            ))
        
        if self.health <= 0:
            if not self._is_already_dead:
                    self._is_already_dead = True
                    self.reports['death'] = f'{self.name} died on turn {self._turn}!'
            self.health = 0
            self._is_dead = True

    def _take_damage(self, damager: 'Pawn', damage: int, damage_type: str, ability=False, ability_name="") -> None:
        barriers = self.effects.get_any_category_name('barrier') # cancels damage
        if barriers:
            barriers[0].on_activate(
                damager=damager,
                total_damage=damage,
                damage_type=damage_type
            )
            return

        parry = self.effects.get_any_category_name('parry') # cancels damage but needs to happen after barrier
        if parry:
            parry[0].on_activate(
                damager=damager,
                total_damage=damage,
                damage_type=damage_type
            )
            return
        
        for effect in self.effects.get_any_category_name('damage_activate'): # only things that don't change the damage but react to it
            effect.on_activate(
                damager=damager,
                total_damage=damage,
                damage_type=damage_type
            )

        if damager is not None:
            # trigger reflect
            self.effects._trigger_reflect(self, damager, damage)

        # damage mitigation due to armor
        damage -= int(round(self.equipment.damage_reduction_percent * damage))

        # trigger vulnerabilities
        if damage_type == "magic":
            increase = sum(
                [effect.take_bonus_damage_percent for effect in self.effects.find_effect_text('magic vulnerability')])
            damage += int(round(increase * damage))
            self.effects.remove_name('magic vulnerability')

        elif damage_type == "poison":
            if self.effects.find_effect_text('poison vulnerability'):
                damage *= 2

        dam_reduce = sum([effect.take_bonus_damage_percent for effect in self.effects.get_any_category_name('modifier')]) # buff/debuffs that affect damage taken on self
        damage += round(dam_reduce*damage)
        for effect in self.effects.get_any_category_name('modifier'):
            effect.on_activate(
                damager=damager,
                total_damage=damage,
                damage_type=damage_type
            )

        # trigger resistances
        resists = sum([effect.take_bonus_damage_percent for effect in self.effects.find_effect_text(f'{damage_type} resist')])
        damage -= int(round(resists * damage))

        # update action log with damage taken
        self.action_history.append(
            Action(
                turn=self._turn,
                type="damage",
                action_name=f"{self.name} took {damage if damage >= 0 else 0} damage from {damager.name if isinstance(damager, Pawn) else 'the tile'}!",
                actor=damager,
                target=self,
                failed=False,
                damage=damage,
                ability_used=ability_name if ability else (damager.current_action.action_name if isinstance(damager, Pawn) else 'Tile'), #type: ignore
                actor_effects=damager.effects.to_dict() if isinstance(damager, Pawn) else None,
                target_effects=self.effects.to_dict()
            ))

        if damage > 0:
            self.health -= damage
            self._was_hit = True

            if self.health <= 0:
                if not self._is_already_dead:
                    self._is_already_dead = True
                    self.reports['death'] = f'{self.name} died on turn {self._turn}!'
                self.health = 0
                self._is_dead = True

    def _heal(self, amount: int, force=False) -> None:
        if not self._is_dead or force:
            self.health += amount
            if self.health > self.health_max:
                self.health = self.health_max

            # alive them if they have health
            if force and self.health > 0:
                self._is_dead = False

    def damage_report(self) -> list[dict[str, Any]]:
        dr = []
        for action in self.action_history:
            if action.damage is not None:
                dr.append(
                    {'turn':action.turn,
                    'damage':action.damage,
                    'ability':action.ability_used,
                    'source':action.actor,
                    'damager_effects':action.actor_effects,
                    'target_effects':action.target_effects}
                )
        return dr

    ###############################
    # ~~~ Passthrough Methods ~~~ #
    ###############################

    def has_effect(self, effect: Union[str, Effect]) -> bool:
        return self.effects.count(effect) > 0

    def equip(self, item: Union[Gear, GearSet]) -> None:
        return self.equipment.equip(item)

    def unequip(self, item: Union[Gear, GearSet, str]) -> None:
        return self.equipment.unequip(item)

    @property
    def poisoned(self) -> bool:
        return self.effects.poisoned

    @property
    def rooted(self) -> bool:
        return self.effects.rooted

    @property
    def stunned(self) -> bool:
        return self.effects.stunned

    @property
    def blinded(self) -> bool:
        return self.effects.blinded

    @property
    def active_effects(self) -> list[str]:
        return [e.name for e in self.effects.active_effects]

    @property
    def vulnerable(self) -> bool:
        return self.effects.vulnerable

    @property
    def vulnerabilities(self) -> list[Effect]:
        return self.effects.vulnerabilities

    def vulnerable_to(self, damage_type: str):
        """
        Returns True if the pawn is vulnerable to the given damage type.

        e.g. 
        `pawn.vulnerable_to('magic')` -> `True`
        """
        return self.effects.vulnerable_to(damage_type)

    @property
    def resistances(self) -> list[Effect]:
        return self.effects.resistances

    def resistant_to(self, damage_type: str):
        """
        Returns True if the pawn is resistant to the given damage type.

        e.g. 
        `pawn.resistant_to('magic')` -> `True`
        """
        return self.effects.resistant_to(damage_type)

    #############################
    # ~~~ Effect Management ~~~ #
    #############################

    # TODO: THIS IS A DUPE I THINK
    def _add_effect(self, effect: Effect) -> None:
        self.effects.add(effect)

    def _tick(self) -> None:
        # apply effects
        for effect in self.effects:
            if effect.new:
                effect.on_create()
                effect.new = False
            effect.on_tick()
            self._tick_damage(effect)

        # TODO: magic vulnerability from fire + frost vuln pair
        frost = self.effects.find_effect_text('frost resistance')
        fire = self.effects.find_effect_text('fire resistance')
        if frost and fire:
            for ice_effect, fire_effect in zip(frost, fire):
                self.effects.add(MagicVulnerability())
                self.effects.remove_one(ice_effect)
                self.effects.remove_one(fire_effect)

        # increment turn counter
        self._turn += 1

    def _post_tick(self) -> None:
        # tick self.effects
        self.effects._tick()  # updates durations and removes expired effects

        # decrement ability cooldowns
        if self._ability_cooldowns:
            for name, cooldown in self._ability_cooldowns.items():
                if cooldown > 0:
                    self._ability_cooldowns[name] -= 1

        # reset action flags
        self.acted_this_turn = False
        self.moved_this_turn = False
        self._was_hit = False
        self.current_action = None

    def stacks(self, effect: Union[str, Effect]) -> int:
        '''
        Return the number of stacks of the specified effect that are currently active on the player.

        e.g. stacks('poison') -> 2
        or   stacks(Poison()) -> 2
        '''
        return self.effects.count(effect)

    @property
    def _marquis(self) -> str:
        # | ❤️❤️❤️❤️❤️❤️🖤🖤🖤 | 🦶 | Oingo Boingo     | 💚2, 💪4, 👀

        health = '❤️' * math.ceil((self.health / self.health_max) * 10)

        empty_health = '🖤' * (10 - len(health)//2)
        action = '🦶' if self.moved_this_turn else '👊'

        return f"| {health}{empty_health} | {action} | {self.symbol}:{self.name:<20} | {self.effects._marquis:<40}"

    ##########################
    # ~~~ Dunder Methods ~~~ #
    ##########################

    def __repr__(self) -> str:
        return f"{self.name} ({clean_name(self.__class__.__name__)}), {self.health}/{self.health_max} HP"

    def __str__(self) -> str:
        return f"{self.name} ({clean_name(self.__class__.__name__)}), {self.health}/{self.health_max} HP"


@dataclass
class Action:
    turn: int
    type: Literal['ability', 'move', 'damage']
    action_name: str
    actor: Union[Pawn, str]
    target: Union[Pawn, Point, str, None]
    failed: bool = field(default=False, init=True)
    failed_reason: str = field(default='', init=True)
    ability_used: Union[str, None] = field(default=None, init=True)
    damage: Union[int, None] = field(default=None, init=True)
    actor_effects: Union[Counter, None] = field(default=None, init=True)
    target_effects: Union[Counter, None] = field(default=None, init=True)

    def __post_init__(self):
        self.action_name = clean_name(self.action_name)
        if isinstance(self.actor, Pawn):
            self.actor = f"{self.actor.name} the {clean_name(self.actor.__class__.__name__)}"
        if isinstance(self.target, Pawn):
            self.target = f"{self.target.name} the {clean_name(self.target.__class__.__name__)}"

    def __repr__(self):
        if self.type == "damage":
            return self.action_name

        message = f"Turn {self.turn}: {self.actor} "

        # if it's an ability or a move
        if self.type == 'ability':
            message += "tried to use " if self.failed else "used "
        elif self.type == 'move':
            message += "tried to move " if self.failed else 'moved '
        elif self.type == 'damage':
            message += "tried to damage " if self.failed else "damaged "

        # if it didn't fail and it's an ability with a target
        if (self.type == 'ability' and self.target is not None):
            message += f"{self.action_name} on {self.target}"
        elif self.type == 'damage':
            message += f"{self.target.name if hasattr(self.target, 'name') else self.target}!" # type: ignore
        # if it didn't fail and it's a move
        elif isinstance(self.target, Point) and self.type == 'move':
            message += f"to {self.target}"

        # if it didn't fail and it's an ability without a target
        else:
            message += f"{self.action_name}"

        # if it failed (continued)
        if self.failed and self.failed_reason:
            message += f", but failed because {self.failed_reason}"
        elif self.failed:
            message += ", but failed!"

        return message

    def __str__(self):
        return self.__repr__()

if __name__ == "__main__":
    p = Pawn(name="Oingo Boingo", position=(0,0), health_max=100)
    print(f"Pawn named {p.name} at position {p.position} with {p.health}/{p.health_max} HP")
    print(p)



import time
import hashlib
from collections import Counter
from typing import Iterator, Union

_SORT_PRIORITY = {
    "Might": 1,
    "Toughness": 2,
    "Frailty": 3,
    "Expose Weakness": 4,
    "Barrier": 5,
    "HoT": 6,
    "DoT": 7,
    "Blind": 8,
    "Stun": 9,
    "Root": 10,
    "Reflect": 11
}


class Effect:
    """
    A class that represents an effect.  

    `name` [str] is the name of the effect.  

    `duration` [int|float("inf")] This indicates the number of ticks that the effect will remain active.
    The default duration is infinite (`float("inf")`).  

    `damage_over_time` [int] This is the amount of damage that the bearer of the effect will take each
    tick. The default value is zero. Negative values will heal the bearer of the effect.

    `bonus_damage` [int] This affects the amount of damage that the bearer of the effect deals. Positive
    values increase damage, negative values decrease damage.

    `bonus_damage_received` [int] This affects the amount of damage that the bearer of the effect will
    receive. Positive values increase the damage that the bearer receives, negative values decrease it.  

    `bonus_movement` [int] is an integer amount of bonus grid squares that the pawn will be able to move
    each tick. See the Haste effect for an example of this. Positive values will increase the number of
    squares that a Pawn can move, and negative values will decrease the number of squares. The default
    value is zero.  

    `bonus_max_health` [int] is the amount of hit points added to the Pawn's maximum hit points when this
    effect is applied. A positive number will increase the Pawn's maximum hit points; a negative number
    will decrease the Pawn's hit points. The default value is zero.
    """

    def __init__(self,
                 name: str,
                 category: set[str] | None = None,
                 duration: Union[int, float] = float("inf"),
                 description: str = "",
                 new: bool = True,
                 symbol: str = "⚙",
                 _stamp: float = time.time(),
                 bonus_movement: int = 0,
                 heal_over_time: int = 0,
                 bonus_max_health: int = 0,
                 bonus_max_health_percent: float = 0,
                 damage_over_time: int = 0,
                 deal_bonus_damage_amount: int = 0,
                 deal_bonus_damage_percent: float = 0,
                 take_bonus_damage_amount: int = 0,
                 take_bonus_damage_percent: float = 0,
                 reflect_damage_amount: int = 0,
                 reflect_damage_percent: float = 0):

        self.name = name
        self.duration = duration
        self.category = category or set()
        self.description = description
        self.new = new
        self.symbol = symbol
        self._stamp = _stamp
        self.bonus_movement: int = bonus_movement
        self.heal_over_time: int = heal_over_time
        self.bonus_max_health: int = bonus_max_health
        self.bonus_max_health_percent: float = bonus_max_health_percent
        self.damage_over_time: int = damage_over_time
        self.deal_bonus_damage_amount: int = deal_bonus_damage_amount
        self.deal_bonus_damage_percent: float = deal_bonus_damage_percent
        self.take_bonus_damage_amount: int = take_bonus_damage_amount
        self.take_bonus_damage_percent: float = take_bonus_damage_percent
        self.reflect_damage_amount: int = reflect_damage_amount
        self.reflect_damage_percent: float = reflect_damage_percent

    def on_create(self, *args, **kwargs):
        ...
    
    def on_expire(self, *args, **kwargs):
        ...
    
    def on_tick(self, *args, **kwargs):
        ...

    def on_activate(self, *args, **kwargs):
        ...

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.duration} turn{'s' if self.duration != 1 else ''})"

    def __str__(self):
        return self.symbol

    def __eq__(self, other):
        if not isinstance(other, Effect):
            raise NotImplemented("Cannot compare Effect to non-Effect object")
        return self.__hash__() == other.__hash__() 

    def __hash__(self) -> int:
        return int.from_bytes(hashlib.md5(''.join((
            self.name, ''.join(self.category), self.description, self.symbol
        )).encode()).digest())


class Effects:
    '''
    Convenience class for a collection of effects.

    `add()`, `remove()` add or remove effects from the collection.

    `tick()` decrements the duration of the effects in the collection.

    `bonus_damage`, `bonus_damage_received`, `bonus_movement`, and`bonus_max_health` are all the collected
    values of the effects in the collection.
    '''

    def __init__(self, *effects: Effect):
        self._effects: list[Effect] = list(effects) if effects else []
        self._reflected = False

    #################################
    # ~~ Tick and trigger methds ~~ #
    #################################

    def _tick(self) -> None:
        'decrement the duration of all effects in the collection'

        self.reflected = False
        for effect in self._effects:
            effect.duration -= 1
            if effect.duration <= 0:
                effect.on_expire()
        self._effects = list(filter(lambda e: e.duration > 0, self._effects))

    def _trigger_reflect(self, damager, target, damage: int) -> None:
        'trigger the effects in the collection that reflect damage'
        self.reflected = True
        for effect in self.get_any_category_name('reflect'):
            target._take_damage(
                damager,
                int(round(damage * effect.reflect_damage_percent +
                    effect.reflect_damage_amount)),
                damage_type='reflect',
                ability=True,
                ability_name="Reflect")

    ########################################
    # ~~ Add, remove, and count effects ~~ #
    ########################################

    def add(self, effect: Effect) -> None:
        'add a single effect to the collection'
        self._effects.append(effect)

    def add_stacks(self, effect_constructor, stacks=1, **kwargs) -> None:
        'add a number of stacks of an effect to the collection'
        for _ in range(stacks):
            self.add(effect_constructor(**kwargs))

    def remove(self, effect: Effect) -> None:
        'remove an effect from the collection'
        self._effects = list(
            filter(lambda e: e != effect, self._effects))

    def remove_category(self, category: str) -> None:
        'remove all effects in the collection with the given category'
        self._effects = list(
            filter(lambda e: category not in e.category, self._effects))

    def remove_name(self, name: str) -> None:
        'remove all effects in the collection with the given name'
        self._effects = list(
            filter(lambda e: e.name.lower() != name.lower(), self._effects))

    def remove_all(self, *effects: Effect) -> None:
        'remove one effect from the collection'
        for effect in effects:
            self.remove(effect)

    # TODO: __TEST THIS__
    def remove_one(self, effect: Effect) -> None:
        'remove one effect from the collection'
        self._effects.pop(self._effects.index(effect))

    def count(self, effect: Union[str, Effect]) -> int:
        'return the number of effects in the collection'
        if isinstance(effect, Effect):
            return len(list(filter(lambda e: e.name == effect.name, self._effects)))
        return len(list(filter(lambda e: e.name.lower() == effect.lower(), self._effects)))

    #############################################
    # ~~~ Convenience properties and methods ~~~#
    #############################################

    @property
    def stunned(self) -> bool:
        'return True if the collection contains a Stun effect'
        return self.count('Stun') > 0

    @property
    def blinded(self) -> bool:
        'return True if the collection contains a Blind effect'
        return self.count('Blind') > 0

    @property
    def rooted(self) -> bool:
        'return True if the collection contains a Root effect'
        return self.count('Root') > 0

    @property
    def poisoned(self) -> bool:
        'return True if the collection contains a Poison effect'
        return self.count('Poison') > 0

    @property
    def vulnerabilities(self) -> list[Effect]:
        'return a list of Vulnerability effects in the collection'
        return [effect for effect in self._effects if 'vulnerab' in effect.name.lower()]

    def vulnerable_to(self, damage_type: str) -> bool:
        'return True if the collection contains a Vulnerability effect of the specified damage type'
        if (vulns := self.vulnerabilities):
            return any(damage_type.lower() in e.name.lower() for e in vulns)
        return False

    @property
    def resistances(self) -> list[Effect]:
        'return a list of Resistance effects in the collection'
        return [effect for effect in self._effects if 'resist' in effect.name.lower()]

    def resistant_to(self, damage_type:str) -> bool:
        'return True if the collection contains a Resistance effect of the specified damage type'
        if (resists := self.resistances):
            return any(damage_type.lower() in e.name.lower() for e in resists)
        return False

    @property
    def vulnerable(self):
        return bool(self.vulnerabilities)

    @property
    def has_active_effects(self):
        return len(self._effects) > 0

    @property
    def active_effects(self) -> list[Effect]:
        return list(set(self._effects))

    #################
    # ~~ Getters ~~ #
    #################

    def get_all_category_name(self, *categories: str) -> list[Effect]:
        'return a list of effects in the collection that have all the specified categories'
        return list(filter(
            lambda e: all(category.lower()
                          in e.category for category in categories),
            self._effects))

    def get_any_category_name(self, *categories: str) -> list[Effect]:
        'return a list of effects in the collection that have the specified category'
        return list(filter(
            lambda e: any(category.lower()
                          in e.category for category in categories),
            self._effects))

    def get_category_effects(self, category: str, ) -> 'Effects':
        'return a new Effects collection of effects in the collection that have the specified category'
        return Effects(*self.get_any_category_name(category))

    def find_effect_text(self, text: str) -> list[Effect]:
        'return a list of effects in the collection that have the specified text'
        return list(filter(lambda e: text.lower() in e.name.lower(), self._effects))

    def find_effect_exact_text(self, text: str) -> Effect | None:
        'return an effect in the collection that has the specified text'
        return next(filter(lambda e: text.lower() == e.name.lower(), self._effects), None)

    #####################################################
    # ~~ Effect type-specific getters and properties ~~ #
    #####################################################

    def get_extra_damage_effects(self) -> list[Effect]:
        'return a list of effects in the collection that cause bearer to take extra damage'
        return [e
                for e in self._effects
                if any(getattr(e, property) != 0
                       for property in [
                    'take_bonus_damage_amount',
                    'take_bonus_damage_percent']
                )]

    @property
    def deal_damage_effects(self) -> list[Effect]:
        'return a list of effects in the collection that cause bearer to deal extra damage'
        return [e
                for e in self._effects
                if any(getattr(e, property) != 0
                       for property in [
                    'deal_bonus_damage_amount',
                    'deal_bonus_damage_percent']
                )]

    @property
    def damage_over_time(self, use=False) -> int:
        if use:
            for effect in [e for e in self._effects if e.damage_over_time != 0]:
                effect.on_activate()
        return sum(effect.damage_over_time for effect in self._effects)

    @property
    def dot(self) -> int:
        'alias for `damage_over_time`'
        return self.damage_over_time

    def get_damage_over_time_effects(self) -> list[Effect]:
        'return a new `Effects` object of effects in the collection that deal damage over time'
        return [effect for effect in self._effects if effect.damage_over_time != 0]

    def get_dot_effects(self) -> list[Effect]:
        'alias for `get_damage_over_time_effects()`'
        return self.get_damage_over_time_effects()

    @property
    def heal_over_time(self, use=False) -> int:
        'total amount of health that will be healed per turn'
        if use:
            for effect in [e for e in self._effects if e.heal_over_time != 0]:
                effect.on_activate()
        return sum(effect.heal_over_time for effect in self._effects)

    @property
    def hot(self) -> int:
        'alias for `heal_over_time`'
        return self.heal_over_time

    def get_heal_over_time_effects(self) -> 'Effects':
        return Effects(*[effect for effect in self._effects if effect.heal_over_time != 0])

    def get_hot_dot_effects(self) -> list[Effect]:
        return [effect for effect in self._effects
                if effect.heal_over_time != 0 or effect.damage_over_time != 0]

    @property
    def bonus_damage_output(self) -> int:
        return sum(effect.deal_bonus_damage_amount for effect in self._effects)

    def get_damage_output_effects(self) -> list[Effect]:
        return [effect for effect in self._effects if effect.deal_bonus_damage_amount != 0]

    @property
    def bonus_movement(self) -> int:
        return sum(effect.bonus_movement for effect in self._effects)

    def get_movement_effects(self) -> list[Effect]:
        return [effect for effect in self._effects if effect.bonus_movement != 0]

    @property
    def bonus_damage_received(self) -> int:
        return sum(effect.take_bonus_damage_amount for effect in self._effects)

    def get_damage_received_effects(self) -> list[Effect]:
        return [effect for effect in self._effects if effect.take_bonus_damage_amount != 0]

    @property
    def bonus_damage_received_percent(self) -> float:
        return sum(effect.bonus_damage_received_percent for effect in self._effects if not effect)

    def get_damage_received_percent_effects(self) -> list[Effect]:
        return [effect for effect in self._effects if effect.take_bonus_damage_percent != 0]

    @property
    def bonus_max_health(self) -> int:
        return sum(effect.bonus_max_health for effect in self._effects)

    def get_max_health_effects(self) -> list[Effect]:
        return [effect for effect in self._effects if effect.bonus_max_health != 0]

    @property
    def _marquis(self) -> str:
        return " ".join([f"{effect.symbol}{self.count(effect):<2}"
                         for effect in sorted(
                             self.active_effects,
                             key=lambda e: _SORT_PRIORITY.get(e.name, 100))
                         ])

    ########################
    # ~~ Dunder methods ~~ #
    ########################

    def to_dict(self):
        return Counter(e.name for e in self._effects)

    def __repr__(self) -> str:
        return f'Effects({self._effects})'

    def __str__(self) -> str:
        return f'Effects({self._effects})'

    def __iter__(self) -> Iterator[Effect]:
        return iter(self._effects)

    def __getitem__(self, index: int) -> Effect:
        return self._effects[index]

    def __len__(self) -> int:
        return len(self._effects)



import math
import re
from dataclasses import dataclass
from functools import singledispatch
from typing import Generator, Tuple, Union

def clean_name(name):
    name = re.sub(r'([a-z])([A-Z])', r'\1 \2', name)
    return name.replace('_', ' ').title()
    
@dataclass
class Point:
    x: int
    y: int

    def to_tuple(self) -> tuple[int, int]:
        return (self.x, self.y)
        
    def __repr__(self)->str:
        return f"({self.x}, {self.y})"

    def __tuple__(self)->tuple[int, int]:
        return (self.x, self.y)

    def __getitem__(self, key):
        return (self.x, self.y)[key]

    def __hash__(self):
        return hash(self.__tuple__())
    
    def __iter__(self):
        yield self.x
        yield self.y

@singledispatch
def distance_between(a: Union[Point,Tuple[int,int]], b: Union[Point, Tuple[int,int]]) -> float:
    if isinstance(a, tuple):
        a = Point(*a)
    if isinstance(b, tuple):
        b = Point(*b)
    return ((a.x - b.x)**2 + (a.y - b.y)**2)**0.5

@distance_between.register
def _(x1:int, y1:int, x2:int, y2:int) -> float:
    return distance_between(Point(x1, y1), Point(x2, y2))

@singledispatch
def behinds(x1:int, y1:int, x2:int, y2:int) -> tuple[Point, Point, Point]:
    x3 = x1 - x2
    y3 = y1 - y2

    xb = x3 + x1
    yb = y3 + y1

    if x1 == x2 and y1 != y2: # Orthog from side
        return Point(xb-1, yb), Point(xb, yb), Point(xb+1, yb)
    elif x1 != x2 and y1 == y2: # Orthog from top/bottom
        return Point(xb, yb+1), Point(xb, yb), Point(xb, yb-1)

    elif xb < x1 and yb < y1: # Diag down left
        return Point(xb, yb+1), Point(xb, yb), Point(xb+1, yb)
    elif xb > x1 and yb < y1:  # Diag down right
        return Point(xb-1, yb), Point(xb, yb), Point(xb, yb+1)
    elif xb < x1 and yb > y1: # Diag up left
        return Point(xb, yb-1), Point(xb, yb), Point(xb+1, yb)
    else: #xb > x1 and yb > y1: Diag up right
        return Point(xb-1, yb), Point(xb, yb), Point(xb, yb-1)

@behinds.register
def _(location: Point, facing: Point) -> tuple[Point, Point, Point]:
    return behinds(location.x, location.y, facing.x, facing.y)

@singledispatch
def angle_between(x1, y1, x2, y2):
    x3 = x2 - x1
    y3 = y2 - y1
    return math.degrees(math.atan2(y3, x3))

@angle_between.register
def _(location: Point, facing: Point):
    return angle_between(location.x, location.y, facing.x, facing.y)

@singledispatch
def angle_behind(x1, y1, x2, y2):
    return angle_between(x2, y2, x1, y1)

@angle_behind.register
def _(location: Point, facing: Point):
    return angle_behind(location.x, location.y, facing.x, facing.y)

def bresenham(origin: Union[Point, tuple[int,int]], destination: Union[Point, tuple[int,int]]) -> Generator[Point, None, None]:
    "Bresenham's Line Algorithm"
    if isinstance(origin, tuple):
        origin = Point(*origin)
    if isinstance(destination, tuple):
        destination = Point(*destination)
    x1, y1 = origin.x, origin.y
    x2, y2 = destination.x, destination.y
    dx = abs(x2 - x1)
    dy = abs(y2 - y1)
    sx = 1 if x1 < x2 else -1
    sy = 1 if y1 < y2 else -1
    err = dx - dy

    while True:
        if x1 == x2 and y1 == y2:
            break
        e2 = 2 * err
        if e2 > -dy:
            err = err - dy
            x1 = x1 + sx
        if e2 < dx:
            err = err + dx
            y1 = y1 + sy
        yield Point(x1, y1)

def get_adjacent_points(position: Union[Point,Tuple[int,int]]) -> list[Point]:
    "get a list of adjacent points to a given point"
    if isinstance(position, tuple):
        position = Point(*position)
    return list(filter(None,
            [Point(position.x + x, position.y + y)
            for x in [-1, 0, 1]
            for y in [-1, 0, 1]
            if not (x == y == 0)]))


from .entities.equipment import Gear, GearSet

__all__ = ['Shield', 'ClothArmor', 'LeatherArmor', 'ChainmailArmor', 'PlateArmor']

class Armor(Gear):
    def __init__(self,
        name: str,
        category: str,
        description: str,
        damage_reduction_percent: float,
        damage_reduction_number: int = 0,
        damage: int = 0,
        damage_type: str = 'physical',
        bonus_damage_output_percent=0,
        bonus_max_health: int = 0,
        bonus_max_health_percent: float = 0
    ) -> None:
        super().__init__(
            name=name,
            category=category,
            description=description,
            damage=damage,
            damage_type = damage_type,
            bonus_damage_output_percent= bonus_damage_output_percent,
            damage_reduction_number=damage_reduction_number,
            damage_reduction_percent=damage_reduction_percent,
            bonus_max_health = bonus_max_health,
            bonus_max_health_percent = bonus_max_health_percent
        )

class Shield(Gear):
    def __init__(self) -> None:
        super().__init__(
            category="offhand",
            name="Shield",
            description="A sturdy shield.",
            damage_reduction_percent=0.05,
        )

class ClothArmor(GearSet):
    def __init__(self) -> None:
        super().__init__(
            name="Cloth Armor",
            description="A simple set of cloth armor.",
        )
        self.gear = [
            Gear(category="head", name="Cloth Hood", damage_reduction_percent=0.01),
            Gear(category="chest", name="Cloth Robe", damage_reduction_percent=0.01),
            Gear(category="legs", name="Cloth Pants", damage_reduction_percent=0.01),
            Gear(category="feet", name="Cloth Shoes", damage_reduction_percent=0.01),
            Gear(category="hands", name="Cloth Gloves", damage_reduction_percent=0.01),
        ]


class LeatherArmor(GearSet):
    def __init__(self) -> None:
        super().__init__(
            name="Leather Armor",
            description="A creaky set of leather armor.",
        )
        self.gear = [
            Gear(category="head", name="Leather Hood", damage_reduction_percent=0.02),
            Gear(category="chest", name="Leather Robe", damage_reduction_percent=0.02),
            Gear(category="legs", name="Leather Pants", damage_reduction_percent=0.02),
            Gear(category="feet", name="Leather Shoes", damage_reduction_percent=0.02),
            Gear(category="hands", name="Leather Gloves", damage_reduction_percent=0.02),
        ]

class ChainmailArmor(GearSet):
    def __init__(self) -> None:
        super().__init__(
            name="Chainmail Armor",
            description="A jingly set of chainmail armor.",
        )
        self.gear = [
            Gear(category="head", name="Chainmail Hood", damage_reduction_percent=0.03),
            Gear(category="chest", name="Chainmail Robe", damage_reduction_percent=0.03),
            Gear(category="legs", name="Chainmail Pants", damage_reduction_percent=0.03),
            Gear(category="feet", name="Chainmail Shoes", damage_reduction_percent=0.03),
            Gear(category="hands", name="Chainmail Gloves", damage_reduction_percent=0.03),
        ]

class PlateArmor(GearSet):
    def __init__(self) -> None:
        super().__init__(
            name="Plate Armor",
            description="A shiny set of plate armor.",
        )
        self.gear = [
            Gear(category="head", name="Plate Hood", damage_reduction_percent=0.04),
            Gear(category="chest", name="Plate Robe", damage_reduction_percent=0.04),
            Gear(category="legs", name="Plate Pants", damage_reduction_percent=0.04),
            Gear(category="feet", name="Plate Shoes", damage_reduction_percent=0.04),
            Gear(category="hands", name="Plate Gloves", damage_reduction_percent=0.04),
        ]

class MithrilArmor(GearSet):
    def __init__(self) -> None:
        super().__init__(
            name="Mithril Armor",
            description="A beautiful set of mithril armor.",
        )
        self.gear = [
            Gear(category="head", name="Mithril Hood", damage_reduction_percent=0.05),
            Gear(category="chest", name="Mithril Robe", damage_reduction_percent=0.05),
            Gear(category="legs", name="Mithril Pants", damage_reduction_percent=0.05),
            Gear(category="feet", name="Mithril Shoes", damage_reduction_percent=0.05),
            Gear(category="hands", name="Mithril Gloves", damage_reduction_percent=0.05),
        ]

class AdamantiteArmor(GearSet):
    def __init__(self) -> None:
        super().__init__(
            name="Adamantite Armor",
            description="A set of hardened adamantite armor.",
        )
        self.gear = [
            Gear(category="head", name="Adamantite Hood", damage_reduction_percent=0.06),
            Gear(category="chest", name="Adamantite Robe", damage_reduction_percent=0.06),
            Gear(category="legs", name="Adamantite Pants", damage_reduction_percent=0.06),
            Gear(category="feet", name="Adamantite Shoes", damage_reduction_percent=0.06),
            Gear(category="hands", name="Adamantite Gloves", damage_reduction_percent=0.06),
        ]

class RuniteArmor(GearSet):
    def __init__(self) -> None:
        super().__init__(
            name="Runite Armor",
            description="A resplendent set of runite armor.",
        )
        self.gear = [
            Gear(category="head", name="Runite Hood", damage_reduction_percent=0.07),
            Gear(category="chest", name="Runite Robe", damage_reduction_percent=0.07),
            Gear(category="legs", name="Runite Pants", damage_reduction_percent=0.07),
            Gear(category="feet", name="Runite Shoes", damage_reduction_percent=0.07),
            Gear(category="hands", name="Runite Gloves", damage_reduction_percent=0.07),
        ]

class DragonArmor(GearSet):
    def __init__(self) -> None:
        super().__init__(
            name="Dragon Armor",
            description="A master-crafted set of dragon armor.",
        )
        self.gear = [
            Gear(category="head", name="Dragon Hood", damage_reduction_percent=0.08),
            Gear(category="chest", name="Dragon Robe", damage_reduction_percent=0.08),
            Gear(category="legs", name="Dragon Pants", damage_reduction_percent=0.08),
            Gear(category="feet", name="Dragon Shoes", damage_reduction_percent=0.08),
            Gear(category="hands", name="Dragon Gloves", damage_reduction_percent=0.08),
        ]

class DemonArmor(GearSet):
    def __init__(self) -> None:
        super().__init__(
            name="Demon Armor",
            description="A set of armor forged from the bones of demons.",
        )
        self.gear = [
            Gear(category="head", name="Demon Hood", damage_reduction_percent=0.09),
            Gear(category="chest", name="Demon Robe", damage_reduction_percent=0.09),
            Gear(category="legs", name="Demon Pants", damage_reduction_percent=0.09),
            Gear(category="feet", name="Demon Shoes", damage_reduction_percent=0.09),
            Gear(category="hands", name="Demon Gloves", damage_reduction_percent=0.09),
        ]

class GodArmor(GearSet):
    def __init__(self) -> None:
        super().__init__(
            name="God Armor",
            description="A set of armor given to mortals by the gods.",
        )
        self.gear = [
            Gear(category="head", name="God Hood", damage_reduction_percent=0.10),
            Gear(category="chest", name="God Robe", damage_reduction_percent=0.10),
            Gear(category="legs", name="God Pants", damage_reduction_percent=0.10),
            Gear(category="feet", name="God Shoes", damage_reduction_percent=0.10),
            Gear(category="hands", name="God Gloves", damage_reduction_percent=0.10),
        ]


from dataclasses import dataclass, field
from functools import singledispatchmethod
from typing import Protocol


class Loot(Protocol):
    name: str
    weight: float
    value: int

@dataclass
class Gear:
    name: str = field(init=True, hash=True)
    category: str = field(init=True, hash=True)
    description: str = field(init=True, default="", hash=False)

    damage: int = field(init=True, default=0, hash=False)
    damage_type: str = field(init=True, default="physical", hash=False)
    
    bonus_damage_output_percent: float = field(
        init=True, default=0, hash=False)

    damage_reduction_number: int = field(init=True, default=0, hash=False)
    damage_reduction_percent: float = field(
        init=True, default=0, hash=False)

    bonus_max_health: int = field(init=True, default=0, hash=False)
    bonus_max_health_percent: float = field(init=True, default=0, hash=False)

    def __repr__(self) -> str:
        if self.damage:
            return f"{self.name} ({self.category}, {self.damage} damage)"
        return f"{self.name} ({self.category}, {self.damage_reduction_percent:.0%} protection)"

class Empty(Gear):
    def __init__(self, category) -> None:
        super().__init__(
            name="Empty",
            category=category,
            description="An empty slot.",
        )

@dataclass
class GearSet:
    name: str = field(init=True, repr=True, hash=True)
    description: str = field(init=True, default="", repr=True, hash=False)
    gear: list[Gear] = field(init=True, default_factory=list, hash=False)

    def __iter__(self):
        return iter(self.gear)

class Equipment:
    def __init__(self, *gear: Gear) -> None:
        self._gear: dict[str, Gear] = {
            "head": Empty("head"),
            "chest": Empty("chest"),
            "legs": Empty("legs"),
            "feet": Empty("feet"),
            "hands": Empty("hands"),
            "weapon": Empty("weapon"),
            "offhand": Empty("offhand"),
        }

        for item in gear:
            self._gear[item.category] = item

    @singledispatchmethod
    def equip(self, item: Gear) -> None:
        self._gear[item.category] = item
    @equip.register
    def _(self, location:str, item: Gear) -> None:
        self._gear[location] = item
    @equip.register
    def _(self, gear_set: GearSet) -> None:
        for item in gear_set:
            self.equip(item)

    @singledispatchmethod
    def unequip(self, item: Gear) -> None:
        self._gear[item.category] = Empty(item.category)
    @unequip.register
    def _(self, location: str) -> None:
        self._gear[location] = Empty(location)
    @unequip.register
    def _(self, gear_set: GearSet) -> None:
        for item in gear_set:
            self.unequip(item)

    @property
    def damage_reduction(self) -> int:
        return sum([item.damage_reduction_number for item in self])

    @property
    def damage_reduction_percent(self) -> float:
        return sum([item.damage_reduction_percent for item in self])

    @property
    def bonus_damage_output(self) -> int:
        return sum([item.damage for item in self])
    
    @property
    def bonus_damage_output_percent(self) -> float:
        return sum([item.bonus_damage_output_percent for item in self])
    
    def __iter__(self):
        return iter(self._gear.values())
    def __repr__(self) -> str:
        return f"Equipment({', '.join([str(item) for item in self])})"
    def __str__(self) -> str:
        return f"Equipment({', '.join([str(item) for item in self])})"
    def __getitem__(self, key: str) -> Gear:
        return self._gear[key]



from .entities.equipment import Gear

class Mace(Gear):
    def __init__(self) -> None:
        super().__init__(
            category="weapon",
            name="Mace",
            description="A heavy mace.",
            damage=5,
            damage_type="physical",
        )

class Sword(Gear):
    def __init__(self) -> None:
        super().__init__(
            category="weapon",
            name="Sword",
            description="A sharp sword.",
            damage=6,
            damage_type="physical",
        )

class Axe(Gear):
    def __init__(self) -> None:
        super().__init__(
            category="weapon",
            name="Axe",
            description="A sharp axe.",
            damage=7,
            damage_type="physical",
        )

class SideKnife(Gear):
    def __init__(self) -> None:
        super().__init__(
            category="offhand",
            name="Side Knife",
            description="A small knife.",
            damage=2,
            damage_type="physical",
        )

class Dagger(Gear):
    def __init__(self) -> None:
        super().__init__(
            category="weapon",
            name="Dagger",
            description="A sharp dagger.",
            damage=4,
            damage_type="physical",
        )

class Staff(Gear):
    def __init__(self) -> None:
        super().__init__(
            category="weapon",
            name="Staff",
            description="A wooden staff.",
            damage=3,
            damage_type="physical",
        )

class Wand(Gear):
    def __init__(self) -> None:
        super().__init__(
            category="weapon",
            name="Wand",
            description="A magical wand.",
            damage=2,
            damage_type="magical",
        )

class ShortBow(Gear):
    def __init__(self) -> None:
        super().__init__(
            category="weapon",
            name="Short Bow",
            description="A short bow.",
            damage=5,
            damage_type="physical",
        )

class Claymore(Gear):
    def __init__(self) -> None:
        super().__init__(
            category="weapon",
            name="Claymore",
            description="A large, two-handed sword.",
            damage=10,
            damage_type="physical",
        )

class Lute(Gear):
    def __init__(self) -> None:
        super().__init__(
            category = "weapon",
            name = "Lute",
            description = "A guitar-like musical instrument.",
            damage = 2,
            damage_type = "physical"
        )

class TreeTrunk(Gear):
    def __init__(self) -> None:
        super().__init__(
            category = "weapon",
            name = "Tree Trunk",
            description = "A large tree, torn from the earth and used as a club.",
            damage = 20,
            damage_type = "physical",
        )


from typing import Tuple, Union
from .pawn import Pawn
from ..utilities.location import Point, bresenham, distance_between


class Square:
    def __init__(self, position: Point, symbol: str = '⬜', impassable: bool = False, is_water: bool = False,
                 is_burning: bool = False, is_lava: bool = False, damage: int = 0, occupant: Union[Pawn, None] = None) -> None:
        self.position = position
        self._base_symbol = symbol
        self._symbol = symbol
        self._temp_symbol = ''
        self._impassable = impassable
        self.is_water = is_water
        self.is_burning = is_burning
        self.is_lava = is_lava
        self.damage = damage
        self.occupant = occupant

    @property
    def name(self) -> str:
        return f"Tile at {self.position}"

    @property
    def impassable(self) -> bool:
        return self._impassable or self.occupied

    @property
    def symbol(self):
        if self.occupied:
            return self.occupant.symbol  # type: ignore
        if self._temp_symbol:
            return self._temp_symbol
        return self._symbol

    def toggle_burning(self, damage: int = 3) -> None:
        if self.is_burning:
            self.is_burning = False
            self._symbol = self._base_symbol
            self.damage = 0
        else:
            self.is_burning = True
            self._symbol = '🟨'
            self.damage = damage

    def toggle_lava(self) -> None:
        if self.is_lava:
            self.is_lava = False
            self._symbol = self._base_symbol
            self.damage = 0
        else:
            self.is_lava = True
            self._symbol = '🟥'
            self.damage = 10000

    def set_temp_symbol(self, symbol: str) -> None:
        self._symbol = symbol

    def clear_temp_symbol(self) -> None:
        self._symbol = ''

    @property
    def occupied(self) -> bool:
        return self.occupant is not None

    def trigger_effect(self) -> None:
        if self.occupied and (self.is_burning or self.is_lava):
            self.occupant._take_damage(self, self.damage, "fire")  # type: ignore

    def place(self, new_occupant: Pawn) -> str:
        '''Place a pawn in the square; returns True if successful.'''
        if self.occupied:
            return "the square is occupied!"
        if self.impassable:
            return "the square is impassable!"
        self.occupant = new_occupant
        return 'success'

    def __repr__(self):
        return self.symbol if not self.occupied else self.occupant.symbol  # type: ignore

    def __str__(self):
        return self.symbol if not self.occupied else self.occupant.symbol  # type: ignore


class Board:
    # assume all levels are square
    def __init__(self, grid: list[list[Square]] | None = None, grid_size: int = 20):
        if grid:
            self.grid_size = len(grid)
            self.grid = grid
        else:
            self.grid_size = grid_size
            self.grid = []

            for y in range(grid_size):
                self.grid.append([])
                for x in range(grid_size):
                    self.grid[y].append(Square(Point(x, y)))

    def _tick(self):
        for row in self.grid:
            for square in row:
                if square.occupied and square.position != square.occupant.position: # type: ignore
                    square.occupant = None
                square.trigger_effect()

    def at(self, position: Union[Point, Tuple[int, int]]) -> Square | None:
        "get square at position (x, y)"

        if isinstance(position, tuple):
            position = Point(position[0], position[1])

        if (0 > position.x >= len(self.grid[0])) or (0 > position.y >= len(self.grid)):
            return None
        return self.grid[position.y][position.x]

    def place(self, pawn: Pawn, position: Point) -> str:
        '''Place a pawn in the square; returns True if successful.'''
        if self.at(position) is None:
            return "the square is out of bounds!"
        return self.at(position).place(pawn) # type: ignore

    
    #################################################################
    # ~~ Getters, Distance Calculations, and Convenience Methods ~~ #
    #################################################################

    @property
    def dangerous_points(self) -> list[Point]:
        "get a list of points that are dangerous"
        return [square.position for row in self.grid for square in row if square.is_burning or square.is_lava]

    @property
    def dangerous_positions(self) -> list[Tuple[int, int]]:
        "get a list of positions that are dangerous"
        return [square.position.to_tuple() for row in self.grid for square in row if square.is_burning or square.is_lava]

    @property
    def impassable_points(self) -> list[Point]:
        "get a list of points that are impassable"
        return [square.position for row in self.grid for square in row if square.impassable]

    @property
    def impassable_positions(self) -> list[Tuple[int, int]]:
        "get a list of positions that are impassable"
        return [square.position.to_tuple() for row in self.grid for square in row if square.impassable]

    def get_squares_at_points(self, *points: Union[Point,tuple[int,int]]) -> list[Square]:
        "get a list of squares at the provided points"
        return list(filter(None, [self.at(point) for point in points]))

    def get_players_in_positions(self, *positions: Union[Point, Tuple[int, int]]) -> list[Pawn]:
        "get a list of players in the provided positions"
        return [self.at(pos).occupant # type: ignore
                for pos in positions
                if self.at(pos) is not None and self.at(pos).occupied] # type: ignore

    def get_players_in_squares(self, *squares: Square) -> list[Pawn]:
        "get a list of players in the provided squares"
        return list(filter(None, [square.occupant for square in squares if square.occupied]))

    def get_players_in_range(self, origin: Union[Point, tuple[int,int]], radius: int) -> list[Pawn]:
        "get a list of players in the provided radius"
        return list(filter(None,
                          [square.occupant
                          for square in self.get_squares_in_radius(origin, radius)]))

    def get_squares_in_radius(self, origin: Union[Point, tuple[int,int]], radius: int) -> list[Square]:
        "get a list of squares in the provided radius"
        if isinstance(origin, tuple):
            origin = Point(*origin)
        return list(filter(None,
                           [self.at(Point(origin.x + x, origin.y + y))
                           for x in range(-radius, radius + 1)
                           for y in range(-radius, radius + 1)
                           if (x ** 2 + y ** 2) <= radius ** 2]))

    def get_adjacent_squares(self, position: Union[Point,tuple[int,int]]) -> list[Square]:
        "get a list of adjacent squares"
        if isinstance(position, tuple):
            position = Point(*position)
        return list(filter(None,
                           [self.at(Point(position.x + x, position.y + y))
                           for x in [-1, 0, 1]
                           for y in [-1, 0, 1]
                           if not (x == y == 0)]))

    def get_adjacent_entities(self, origin: Pawn) -> list[Pawn]:
        "get a list of all entities in melee range of the origin pawn"
        return self.get_players_in_squares(*self.get_adjacent_squares(origin.position))
    
    def get_melee_range_entities(self, origin: Pawn) -> list[Pawn]:
        "get a list of all entities in melee range of the origin pawn"
        return self.get_adjacent_entities(origin)

    def get_nearest_player_to(self, origin: Pawn) -> Pawn:
        "get the nearest player to the origin pawn"
        player = None
        radius = 1
        while player is None:
            players = self.get_players_in_range(origin.position, radius)
            if players:
                player = min(players, key=lambda p: distance_between(origin, p)) # type: ignore
            radius += 1
        return player

    def distance_between(self, origin: Union[Pawn, Point], destination: Union[Pawn, Point]) -> float:
        "get the distance between two points"
        if isinstance(origin, Pawn):
            origin = origin.position
        if isinstance(destination, Pawn):
            destination = destination.position
        return distance_between(origin, destination)

    def get_squares_in_line(self, origin: Point, destination: Point) -> list[Square]|None:
        if self.at(origin) is not None and self.at(destination) is not None and origin != destination:
            return self.get_squares_at_points(*list(bresenham(origin, destination)))

    def __repr__(self):
        return f"Board({len(self.grid)} * {len(self.grid[0])} grid)"

    def __str__(self):
        s = ""
        for y in range(len(self.grid)-1, -1, -1):
            for x in range(len(self.grid[0])):
                s += str(self.grid[y][x].symbol)
            s += "\n"

        return s

    def __getitem__(self, position: Union[Point,tuple[int,int]]) -> Square | None:
        return self.at(position)




from .entities.board import Board
from .entities.characters import Party
from .bosses import Boss, TrainingDummy
from ._level import Level
from .utilities.map_making import get_map


class BlankLevel(Level):
    def __init__(self, party: Party, boss: Boss, board: Board, show_board: bool=True, tick_speed: float=0.25) -> None:
        super().__init__(board=board, party=party, boss=boss, show_board=show_board, tick_speed=tick_speed)

class MovementTraining(Level):
    def __init__(self, party: Party, boss:Boss, show_board: bool=True, tick_speed: float=0.25) -> None:
        board = get_map('simple_map.json')
        super().__init__(board=board, party=party, boss=boss, show_board=show_board, tick_speed=tick_speed)

class ForestPath(Level):
    def __init__(self, party: Party, boss:Boss, show_board: bool=True, tick_speed: float=0.25) -> None:
        board = get_map('forest_path.json')
        super().__init__(board=board, party=party, boss=boss, show_board=show_board, tick_speed=tick_speed)
        
class RiverFord(Level):
    def __init__(self, party: Party, boss:Boss, show_board: bool=True, tick_speed: float=0.25) -> None:
        board = get_map('river_ford.json')
        super().__init__(board=board, party=party, boss=boss, show_board=show_board, tick_speed=tick_speed)


import time
import random
import os
import sys

try:
    from IPython.display import clear_output
except ModuleNotFoundError:
    print("IPython not found, defaulting to system clear-screen command to clear output.",
    "This causes the game to flicker in most cases; for the best experience use Jupyter Notebook or Jupyter Lab.", file=sys.stderr)
    def clear_output(wait: bool=False):
        if not wait:
            os.system('cls' if os.name == 'nt' else 'clear')
        else:
            sys.stdout.write('\b'*5000)

from .entities.board import Board
from .entities.characters import Party
from .entities.pawn import Pawn

from .utilities.location import Point

from .bosses import Boss

class Level:
    board: Board
    party: Party
    boss: Boss
    turn_count: int

    def __init__(self, board: Board, party: Party, boss: Boss, show_board: bool = True, tick_speed: float = 0.25):
        party_starty = [square
                        for row in board.grid
                        for square in row
                        if square.symbol == '🟢']
        boss_starty = [square for row in board.grid for square in row if square.symbol == '🔴'][0]
        
        for pawn, square in zip(party, party_starty):
            pawn._position = square.position
            pawn.move_history = [square.position]
            pawn.face(random.choice(board.get_adjacent_squares(pawn.position))) # type: ignore

        boss._position = boss_starty.position
        boss.move_history = [boss_starty.position]
        boss.face(random.choice(board.get_adjacent_squares(boss.position))) # type: ignore
        
        self.board = board
        self.party = party
        self.boss = boss
        self.turn_count = 0
        self.show_board = show_board
        self.tick_speed = tick_speed
        
        # place the pawns
        for pawn in self.party:
            self.board.place(pawn, pawn.position)
        self.board.place(self.boss, self.boss.position)

    def move(self, pawn: Pawn, position: Point):
        if pawn.moved_this_turn:
            result = self.board.place(pawn, position)
            if result != 'success':
                pawn._revert_position(result)

    @property
    def _marquis(self):
        return f"~~~~~~ TURN {self.turn_count:<4}~~~~~~\n{self.party._marquis}\n{'_'*80}\n{self.boss._marquis}"

    def __iter__(self):
        while self.party.is_alive and self.boss.is_alive:

            # check movements
            for player in self.party:
                self.move(player, player.position)
            
            if self.turn_count:
                self.boss._tick()
                self.boss._tick_logic(self.party, self.board)
                self.move(self.boss, self.boss.position)
                self.boss._post_tick()
            else:
                self.boss._post_tick()

            # send tick to party, boss
            self.party._tick()
            self.board._tick()

            self.turn_count += 1

            if self.show_board:
                clear_output(wait=True)
                print(self)
                time.sleep(self.tick_speed)
            self.party._post_tick()

            # player turns happen here
            yield self.turn_count
        
        # game over
        for player in self.party:
            self.move(player, player.position)
        self.boss._tick()
        self.party._tick()
        self.board._tick()
        
        clear_output(wait=True)
        print(self)

        self.party._post_tick()
        
        yield self.turn_count

    def __str__(self):
        return f"{self._marquis}\n{self.board}"

class DummyGame(Level):
    def __init__(self, board: Board, party: Party, boss: Boss, show_board: bool = True, tick_speed: float = 0.25):
        super().__init__(board, party, boss, show_board, tick_speed)




from typing import Tuple, Union
from .equipment import GearSet
from .pawn import Pawn, Action, _action_decorator
from .effects import Effect
from ..utilities.location import Point
from ..armor import ClothArmor

class Character(Pawn):
    def __init__(self, name: str, symbol: str, role: str, position: Union[Point, Tuple[int, int]] = Point(0, 0), health_max: int = 100, gear: GearSet = ClothArmor()) -> None:
        super().__init__(name, position, health_max, symbol, gear=gear)
        self.name = name
        self.role = role

    @property
    def last_action(self) -> Action | None:
        if self.action_history:
            return self.action_history[-1]

    @property
    def last_successful_action(self) -> Action | None:
        for action in reversed(self.action_history):
            if not action.failed:
                return action

    def __repr__(self):
        return f"Character({self.name}, {self.position}, {self.health}/{self.health_max}, {self._symbol})"

    def __str__(self):
        return f"Character({self.name}, {self.position}, {self.health}/{self.health_max}, {self._symbol})"

class Party:
    def __init__(self, *members: Character) -> None:
        self.members = members

        # check that they don't have more than one of one type of class
        # for member in members:
        #     for other in members:
        #         if (member is not other and
        #         member.__class__.__name__ == other.__class__.__name__):
        #             raise ValueError("Party must be made up of unique classes")

        if len((walrus := list(filter(lambda x: x.role.lower() == 'tank', members)))) != 1:
            raise ValueError("Party must have exactly one tank")
        self._tank = walrus[0]

        if len((walrus := list(filter(lambda x: x.role.lower() == 'healer', members)))) != 1:
            raise ValueError("Party must have exactly one healer")
        self._healer = walrus[0]

        if len((walrus := list(filter(lambda x: x.role.lower() == 'dps', members)))) != 2:
            raise ValueError("Party must have at least two dps")
        self._dps = tuple(walrus)
        self.name = "The party"
        self.position = Point(0, 0)

    def _tick(self):
        for member in self.members:
            member._tick()

    def _post_tick(self):
        for member in self.members:
            member._post_tick()

    def _add_effect(self, effect: Effect):
        for member in self.members:
            member.effects.add(effect)

    def closest_to(self, target: Pawn) -> Character:
        return min(self.members, key=lambda x: x.distance_from(target))

    def furthest_from(self, target: Pawn) -> Character:
        return max(self.members, key=lambda x: x.distance_from(target))

    #############################
    # ~~ Convenience Methods ~~ #
    #############################

    @property
    def lowest_health(self) -> Character:
        return min(self.members, key=lambda x: x.health)

    @property
    def lowest_health_percent(self) -> Character:
        return min(self.members, key=lambda x: x.health_percent)

    @property
    def highest_health(self) -> Character:
        return max(self.members, key=lambda x: x.health)
    
    @property
    def highest_health_percent(self) -> Character:
        return max(self.members, key=lambda x: x.health_percent)

    @property
    def tank(self) -> Pawn:
        return self._tank

    @property
    def healer(self) -> Pawn:
        return self._healer

    @property
    def dps(self) -> tuple[Pawn, ...]:
        return self._dps

    @property
    def is_alive(self) -> bool:
        return any(member.is_alive for member in self.members)

    @property
    def is_dead(self) -> bool:
        return not self.is_alive

    @property
    def _marquis(self):
        return "\n".join(member._marquis for member in self.members)

    def __repr__(self):
        return f"Party({self.members})"

    def __str__(self):
        return f"Party({self.members})"

    def __iter__(self):
        return iter(self.members)

    def __getitem__(self, key):
        attr = [member
                for member in self.members
                if member.name.lower() == key.lower()
                or key.lower() in member.__class__.__name__.lower()]
        return attr[0] if attr else None


class DummyHero(Character):
    def __init__(self, name: str, position: Union[Point, Tuple[int, int]], role, symbol:str='🍄') -> None:
        super().__init__(name, symbol, role, position, 100)
        self.equip(ClothArmor())
        self.turn = None

    @_action_decorator(cooldown=2, melee=True) #type: ignore
    def ball_punch(self, target: Pawn) -> None:
        target._take_damage(self, 100, "spiritual")
        target.effects.add(Effect(name="Dummy Hero Ball Punch", damage_over_time=2, duration=5))


from typing import Tuple, Union
from .effects import Effects
from .pawn import Pawn
from .characters import Party

from ..utilities.location import Point, distance_between


import functools

class Monster(Pawn):
    # type: ignore
    def __init__(self, name: str, position: Union[Point, Tuple[int, int]] = Point(0, 0), health_max: int = 500, symbol: str = '👹'):
        super().__init__(name=name, position=position, health_max=health_max, symbol=symbol)
        self.effects = Effects()
        self.health = self.health_max
        self.telegraph: Union[str,None] = None

    def _get_target(self, party: Party) -> Pawn:
        '''Get the target to attack. Tank -> DPS -> Healer.'''
        if party.tank.is_alive:
            return party.tank
        elif any(member.is_alive for member in party.dps):
            nearest = None
            distance = 9999
            for member in party.dps:
                if distance_between(self.position, member.position) < distance and member.is_alive:
                    nearest = member
                    distance = distance_between(self.position, member.position)
            return nearest # type: ignore
        else:
            return party.healer

        
# vulnerabilities: logic to check effects and double the values if they're in the "vulnerable" category

#TODO: resistances and vulnerabilities

#TODO: telegraph system
#TODO: aggro system! "I'm looking at _____(pawn)"
#     takes a function like a callback to decide?
#     takes a list of pawns?
#     *   checks pawns in a radius around the monster?
#     *   checks pawn health?




import copy
import abc
import heapq
import itertools
import random
from typing import List, Tuple, Union

from .entities.board import Board
from .entities.monster import Monster
from .entities.pawn import Pawn, _action_decorator
from .entities.characters import Party

from .utilities.location import Point, distance_between, bresenham

from .debuffs import Curse, Embarrassed, Frailty, Stun
from .buffs import HoT
from .weapons import Claymore, Dagger, Sword, TreeTrunk

class Boss(Monster, abc.ABC):
    @abc.abstractmethod
    def _tick_logic(self, party: Party, board: Board):
        ...

    def _astar(self, board: Board, start: Union[Point,Tuple], goal: Union[Point,Tuple]) -> list[Point] | None:

        start = tuple(start)
        goal = tuple(goal)

        neighbors = [(0, 1), (0, -1), (1, 0), (-1, 0),
                    (1, 1), (1, -1), (-1, 1), (-1, -1)]
        close_set = set()
        came_from = {}
        gscore = {start: 0}
        fscore = {start: distance_between(start, goal)}

        oheap = []
        heapq.heappush(oheap, (fscore[start], start))
        while oheap:
            current = heapq.heappop(oheap)[1]
            if current == goal:
                data = []
                while current in came_from:
                    data.append(Point(*current))
                    current = came_from[current]
                return data[::-1]

            close_set.add(current)

            for i, j in neighbors:
                neighbor = current[0] + i, current[1] + j
                tentative_g_score = gscore[current] + distance_between(current, neighbor)

                if 0 <= neighbor[0] < board.grid_size:
                    if 0 <= neighbor[1] < board.grid_size:
                        if (board.at(neighbor).impassable or board.at(neighbor).is_lava) and neighbor != goal: # type: ignore
                            continue

                    else:
                        # grid bound y walls
                        continue

                else:
                    # grid bound x walls
                    continue

                if neighbor in close_set and tentative_g_score >= gscore.get(neighbor, 0):
                    continue

                if tentative_g_score < gscore.get(neighbor, 0) or neighbor not in [i[1]for i in oheap]:
                    came_from[neighbor] = current
                    gscore[neighbor] = tentative_g_score # type: ignore
                    fscore[neighbor] = tentative_g_score + distance_between(neighbor, goal)

                    heapq.heappush(oheap, (fscore[neighbor], neighbor))



###################
# ~~ TEST BOSS ~~ #
###################

class Golem(Boss):
    '''
    The Golem is a high-damage, implacable foe. It has a large health pool,
    and will steadily attack the party until it is defeated.
    '''
    def __init__(self):
        name="Thunk"
        position=Point(0, 0)
        health_max=20000
        super().__init__(name=name, position=position, health_max=health_max)

    def get_target(self, party: Party) -> Pawn:
        return self._get_target(party)

    def _tick_logic(self, party: Party, board: Board):
        target = self.get_target(party)

        # Telegraphing
        if self._ability_cooldowns.get('Aoe', 0) == 1:
            self.telegraph = "I'm going to attack EVERYONE next turn!"

        # Combat logic
        if distance_between(self.position, target.position) > 1.5:
            path = self._astar(board=board, start=self.position, goal=target.position)
            if path is not None:
                self.move(path[0])
        else:
            if self.is_on_cooldown("aoe"):
                self.attack(target)
            else:
                self.aoe(party)

    @_action_decorator(cooldown=2, melee=True) # type: ignore
    def attack(self, target: Pawn):
        target._take_damage(self, 20, "physical")

    @_action_decorator(cooldown=10) # type: ignore
    def aoe(self, party: Party):
        for pawn in party.members:
            pawn._take_damage(self, 50, "physical")


##################################
# ~~ Training Scenario Bosses ~~ #
##################################

# ~~ First Scenario Boss ~~ #

class TrainingDummy(Boss):
    def __init__(self):
        name="Training Dummy"
        position=Point(0, 0)
        health_max=10000
        super().__init__(name=name, position=position, health_max=health_max)

    def get_target(self, party: Party) -> Pawn:
        return party.closest_to(self)

    @_action_decorator(cooldown=2, melee=False, affected_by_blind=False) # type: ignore
    def shout_at(self, target: Pawn):
        target._add_effect(Embarrassed())

    @_action_decorator(cooldown=3, melee=False, affected_by_blind=False) # type: ignore
    def make_a_ruckus(self, party: Party):
        for pawn in party.members:
            pawn._add_effect(Embarrassed())

    def _tick_logic(self, party: Party, board: Board):
        target = self.get_target(party)

        self.telegraph = f"I'm going to insult the closest person!"

        self.face(target)
        if not self.is_on_cooldown("make a ruckus"):
            self.make_a_ruckus(party)
        else:
            self.shout_at(target)


class LostKobold(Boss):
    def __init__(self):
        name="Lost Kobold"
        position=Point(0, 0)
        health_max=5000
        super().__init__(name=name, position=position, health_max=health_max)
        self.equip(Dagger())

    def get_target(self, party: Party) -> Pawn:
        return self._get_target(party)

    def _tick_logic(self, party: Party, board: Board):
        target = self._get_target(party)

        if distance_between(self.position, target.position) > 1.5:
            path = self._astar(board=board, start=self.position, goal=target.position)
            if path is not None:
                self.move(path[0])
        else:
            self.fumbling_attack(target)

    @_action_decorator(melee=True) # type: ignore
    def fumbling_attack(self, target: Pawn):
        target._take_damage(self, self.calculate_damage(5, target), "physical")


class KoboldMother(Boss):
    def __init__(self):
        name="Kobold Mother"
        position=Point(0, 0)
        health_max=10000
        super().__init__(name=name, position=position, health_max=health_max)
        self.equip(Sword())

    def get_target(self, party: Party) -> Pawn:
        return self._get_target(party)

    def _tick_logic(self, party: Party, board: Board):
        target = self._get_target(party)

        if distance_between(self.position, target.position) > 1.5:
            path = self._astar(board=board, start=self.position, goal=target.position)
            if path is not None:
                self.move(path[0])
        else:
            self.motherly_love(target)

    @_action_decorator(melee=True) # type: ignore
    def motherly_love(self, target: Pawn):
        target._take_damage(self, self.calculate_damage(40, target), "physical")


class KoboldQueen(Boss):
    def __init__(self):
        name="Kobold Queen"
        position=Point(0, 0)
        health_max=20000
        super().__init__(name=name, position=position, health_max=health_max)
        self.equip(Sword())

    def get_target(self, party: Party) -> Pawn:
        return self._get_target(party)

    def _tick_logic(self, party: Party, board: Board):
        target = self._get_target(party)

        if distance_between(self.position, target.position) > 1.5:
            path = self._astar(board=board, start=self.position, goal=target.position)
            if path is not None:
                self.move(path[0])
        else:
            if not self.is_on_cooldown("Curse"):
                self.curse(party)
            else:
                self.savage_strike(target)

    @_action_decorator(melee=True) # type: ignore
    def savage_strike(self, target: Pawn):
        target._take_damage(self, 60, "physical")

    @_action_decorator(cooldown=10) # type: ignore
    def curse(self, party: Party):
        for pawn in list(party.dps) + [party.healer]:
            pawn.effects.add(Curse(caster=self, target=pawn, duration=8, dot_amount=self.calculate_damage(3, pawn)))

class KoboldGoddess(Boss):
    def __init__(self):
        name="Kobold Goddess"
        position=Point(0, 0)
        health_max=20000
        super().__init__(name=name, position=position, health_max=health_max)
        self.equip(Claymore())

        self.last_party_positions : List = [] # #<== Death touch mechanic property

    def get_target(self, party: Party) -> Pawn:
        return self._get_target(party)

    def _tick_logic(self, party: Party, board: Board):
        self.telegraph = None
        target = self._get_target(party)

        if self._turn and self._turn % 5 == 0:
            self.last_party_positions = [p.position for p in party]
            self.telegraph = "I'm going to devour your souls if you don't move next turn!!!"
        elif self.telegraph:
            self.devour_souls(party)
        elif distance_between(self.position, target.position) > 1.5:
            path = self._astar(board=board, start=self.position, goal=target.position)
            if path is not None:
                self.move(path[0])
        else:
            if not self.is_on_cooldown("Curse"):
                self.curse(party)
            else:
                self.savage_strike(target)

    @_action_decorator(affected_by_blind=False) # type: ignore
    def devour_souls(self, party: Party):
        for p in party:
            if p.position in self.last_party_positions:
                p._take_damage(self, 666, "physical")
        self.telegraph = None

    @_action_decorator(melee=True) # type: ignore
    def savage_strike(self, target: Pawn):
        target._take_damage(self, 60, "physical")

    @_action_decorator(cooldown=10) # type: ignore
    def curse(self, party: Party):
        for pawn in list(party.dps) + [party.healer]:
            pawn.effects.add(Curse(caster=self, target=pawn, duration=8, dot_amount=self.calculate_damage(3, pawn)))


# 
# Savage Mountain Troll

# Core Mechanic: Regenerates Health
# Attack: Normal heavy attack against melee target.
# cycle:
    # Mechanic 1: Every 10 turns will attack with extreme damage that applies Stun to the target for 2 turns.
    # Mechanic 2: (Follows Mechanic 1) Let's out a roar that applies 10 stacks of Frailty to Party for 3 turns
    # Mechanic 3: (Follows Mechanic 2) Jumps into the air and smashes down causing heavy damage to entire Party.
# random
    # Mechanic 4: (Telegraphed) Throws a giant rock randomly at furthest target. Anyone at that position will receive massive damage and applies Stun for 2 turns.
# Mechanic 5: Death charges the closest target dealing extreme damage if no target is in melee range for 2 turns (can only occur after turn 13) 

class SavageMountainTroll(Boss):
    def __init__(self):
        name="Morgoth Trollkin"
        position=Point(0, 0)
        health_max=20000
        super().__init__(name=name, position=position, health_max=health_max)
        
        self.equip(TreeTrunk())
        self._add_effect(HoT(name="Trollkin Regeneration", duration=float('inf'), heal_amount=5))

        self._throwing = False
        self._was_in_melee = False
        self._melee_turn_counter = 0
        self._cycle_turn_counter = 0
        self._cycling = False
        self._action_cycle = itertools.cycle([self.decimate, self.roar_of_giants, self.colossal_smash])
        self._furthest_position : Point = Point(0, 0)

    def get_target(self, party: Party) -> Pawn:
        # get closest party member
        return min(party.members, key=lambda p: distance_between(self.position, p.position))

    def _tick_logic(self, party: Party, board: Board):
        self.telegraph = None
        target = self.get_target(party)

        if self._turn and self._turn % 10 == 0:
            self._cycling = True

        if self._throwing:
            self.throw_boulder(self._furthest_position, party)
        
        if self._was_in_melee and distance_between(self.position, target.position) > 1.5:
            self._melee_turn_counter += 1

        if not self.acted_this_turn:
            if self._melee_turn_counter >= 2:
                self.death_charge(party=party, board=board)

            elif self._cycling:
                if self._cycle_turn_counter >= 3:
                    self._cycle_turn_counter = 0
                    self._cycling = False
                else:
                    self._cycle_turn_counter += 1
                    next(self._action_cycle)(party, target=target)
            else:
                path = list(self._astar(board=board, start=self.position, goal=target.position)) # type: ignore
                if path:
                    self.move_toward(path[0])
        
        self._throwing = random.random() <= 0.1 # 10% chance to throw a boulder
        if self._throwing:
            person = max(party.members, key=lambda player: distance_between(self.position, player.position))
            self._furthest_position = copy.copy(person.position)
            self.telegraph = f"The savage mountain troll is about to throw a boulder at {person.name}'s position!"

    @_action_decorator(cooldown=10, melee=True) # type: ignore
    def decimate(self, party: Party, **kwargs):
        target = self.get_target(party)
        target._take_damage(self, self.calculate_damage(90, target), "physical") # high dmg due to equipped TreeTrunk
        self._was_in_melee = True
    
    @_action_decorator(cooldown=10, melee=False, affected_by_blind=False) # type: ignore
    def roar_of_giants(self, party: Party, **kwargs):
        for pawn in party:
            pawn.effects.add_stacks(Frailty, stacks=10, duration=3)
    
    @_action_decorator(cooldown=10, melee=False, affected_by_blind=True, affected_by_root=True) # type: ignore
    def colossal_smash(self, party: Party, **kwargs):
        for pawn in party:
            pawn._take_damage(self, self.calculate_damage(60, pawn), "physical")
    
    @_action_decorator(cooldown=10, melee=False, affected_by_blind=True, affected_by_root=False) # type: ignore
    def throw_boulder(self, point: Point, party: Party):
        unlucky = list(filter(lambda target: target.position == point, party.members))
        if unlucky:
            target = unlucky[0]
            target._take_damage(self, self.calculate_damage(150, target), "physical")
            target.effects.add(Stun(duration=2))
    
    @_action_decorator(cooldown=1, melee=False, affected_by_blind=True, affected_by_root=True) # type: ignore
    def death_charge(self, party: Party, board: Board):
        target = self.get_target(party)
        if distance_between(self.position, target.position) > 1.5:
            path = self._astar(board=board, start=self.position, goal=target.position)
            if path is not None:
                self._teleport(path[-2])
        target._take_damage(self, self.calculate_damage(1000, target), "physical")
        self._was_in_melee = True
        self._melee_turn_counter = 0



import random
from .entities.effects import Effect
from .utilities.location import bresenham

#############################
# ~~ Training Dummy Debuff ~~#
#############################


class Embarrassed(Effect):
    def __init__(self) -> None:
        super().__init__(
            name="Embarrassed",
            duration=3,
            category={'embarrassed', 'debuff', 'spirit', 'spiritual', 'mental', 'curable'},
            description=f'The training dummy shouted, "{self.get_random_insult()}"',
            symbol='✨')

    def get_random_insult(self) -> str:
        insults = [
            "What a pillock!",
            "You absolute berk!",
            "I would pity you if you weren't such a buffoon!",
            "Your nincompoopery knows no bounds!",
            "You absolute ninnyhammer!",
            "You absolute ninny!",
            "You milquetoast!",
            "You pettifogging hobbledehoy!",
            "If you weren't such a cow-handed fopdoodle, I might be scared!",
            "You fragrant man-swine!",
            "If I wanted a kiss I'd have called your mother!",
            "If laughter is the best medicine, your face is curing the world!",
            "If I had a face like yours I'd sue my parents!",
            "If I wanted to kill myself I'd climb your ego and jump to your IQ!",
            "I don't know what makes you so stupid, but it really works!",
            "Calling you an idiot would be an insult to all stupid people!",
            "When your mother dropped you off at the schoolhouse she was fined for littering!",
            "Some babies were dropped on their heads but you were clearly thrown at a wall!",
            "I would slap you, but that would be animal abuse!",
            "If being ugly is a crime, you should be locked up for life!",
            "I don't know what your problem is, but I'll bet it's hard to pronounce!",
            "You are the reason the gene pool needs a lifeguard!",
            "Unfortunately stupidity is not a crime, so you're free to go!",
            "How did you get here, did someone leave your cage open?",
            "Don't you have a terribly empty feeling... in your skull?",
            "Are you always this stupid, or is today a special occasion?",
            "I would agree with you if I wanted us both to be wrong!",
            "Some cause happiness wherever they go. You cause happiness *whenever* you go!",
            "I'm glad to see you're not letting education get in the way of your ignorance!",
            "Don't be ashamed of who you are, that's your parents' job!",
            "You've got two brain cells that are both fighting for third place!",
            "You couldn't pour water out of a boot if the instructions were on the heel!",
            "You are proof God makes mistakes!",
            "Calling you an imbecile is an insult to imbeciles everywhere!",
            "I love what you've done with your hair... how do you get it to come out of the nostrils like that?",
            "If you spend word for word with me, I shall make your wit bankrupt!",
            "You may look like an idiot and talk like an idiot, but don't let that fool you, you really are an idiot!",
            "You have no enemies, but you are intensely disliked by your friends!",
            "You are impossible to underestimate!",
            "As an outsider, what is your perspective on intelligence?",
            "The bar is so low it's practically a tripping hazard in hell, but here you are dancing the limbo with the devil!",
            "If my dog looked like you, I'd shave his backside and teach him to walk backwards!",
            "Some day you'll go far, and we all hope you stay there!",
            "Mirrors cannot talk. Luckily for you, they can't laugh, either!",
            "Your face makes onions cry!",
            "Your teeth are so bad you can eat an apple through a fence!"
            "I'll never forget the first time we met, although I'll keep trying!",
            "You have miles to go before you reach mediocre!",
            "I would prefer a battle of wits, but I see you are unarmed!"
        ]
        return random.choice(insults)


class Poison(Effect):
    def __init__(self, target, duration:int=3, dot_amount:int=3) -> None:
        self._dot_amount = dot_amount
        self.target = target
        super().__init__(name="Poison", duration=duration,
                         category={'dot', 'poison', 'debuff', 'damage', 'damage over time', 'curable'}, symbol='🐍')
        
    @property
    def damage_over_time(self) -> int:
        if self.target.poisoned:
            return self._dot_amount * 2
        return self._dot_amount
    
    @damage_over_time.setter
    def damage_over_time(self, value) -> None:
        self._dot_amount = value


class Stun(Effect):
    def __init__(self, duration) -> None:
        super().__init__(name="Stun", duration=duration,
                         category={'stun', 'debuff', 'debilitating', 'physical', 'curable'}, symbol='💫')


class Root(Effect):
    def __init__(self, duration) -> None:
        super().__init__(name="Root", duration=duration,
                         category={'root', 'slow', 'debilitating', 'debuff', 'curable'}, symbol='🐌')


class Blind(Effect):
    def __init__(self, duration) -> None:
        super().__init__(name="Blind", duration=duration,
                         category={'blind', 'debilitating', 'debuff', 'curable'}, symbol='👀')


class Frailty(Effect):
    def __init__(self, duration) -> None:
        super().__init__(name="Frailty", duration=duration, deal_bonus_damage_percent=-.05,
                         category={'frailty', 'physical', 'debuff', 'damage reduction'}, symbol='🤏')


class ExposeWeakness(Effect):
    def __init__(self, duration) -> None:
        super().__init__(name="Expose Weakness", duration=duration, take_bonus_damage_percent=.05,
                         category={'expose weakness', 'physical', 'debuff', 'modifier'}, symbol='🥴')


class PoisonVulnerability(Effect):
    def __init__(self, duration) -> None:
        super().__init__(name="Poison Vulnerability", duration=duration,
                         category={'poison', 'vulnerable', 'debuff', 'curable'}, symbol='🤢')


class MagicVulnerability(Effect):
    def __init__(self) -> None:
        super().__init__(name="Magic Vulnerability", duration=10, category={
            'magic', 'magic vulnerability', 'debuff', 'vulnerable', 'curable'}, take_bonus_damage_percent=.10, symbol='🤩')


class DoT(Effect):
    def __init__(self, target, name: str, duration:int=3, dot_amount:int=3) -> None:
        self.target=target
        super().__init__(name=name, duration=duration, damage_over_time=dot_amount,
                         category={'dot', 'debuff', 'damage', 'damage over time', 'curable'}, symbol='🩸')


class Curse(Effect):
    def __init__(self, caster, target, duration, dot_amount) -> None:
        super().__init__(name="Curse", duration=duration, damage_over_time=dot_amount,
                         category={'dot', 'debuff', 'damage', 'damage over time', 'curable'}, symbol='🩸')
        self.target = target
        self.caster = caster

    def on_expire(self):
        self.target._take_damage(self.caster, 10000, 'curse', True, "Curse")

class Doom(Effect):
    def __init__(self, caster, target, damage: int) -> None:
        self.caster = caster
        self.base_damage = damage
        self.multiplier = 1
        self.triggered = 0
        super().__init__(name="Doom", description="Your doom is near", duration=15, category={'doom', 'curable'}, symbol='☠️')
        self.target=target
    
    def increase_doom(self):
        self.triggered += 1
        
        for _ in self.target.effects.get_any_category_name('debuff'):
            self.multiplier += 0.05
        self.target.effects.remove_category('debuff')

        self.base_damage *= self.multiplier
        self.duration = 15

        if self.triggered >= 6:
            self.duration = 0

    def on_expire(self):
        self.duration=0
        self.target._take_damage(self.caster, int(self.base_damage), 'spirit', True, "Doom")

class ShadeOfDeath(Effect):
    def __init__(self, caster, target, damage:int=0) -> None:
        self.position = caster.position
        self.damage_total = damage
        self.target = target
        self.caster = caster
        super().__init__( name='Shade of Death', category={'debuff', 'damage'}, symbol='👻')
        self.duration = len(list(bresenham(self.position, target.position)))

    @property
    def duration(self) -> int:
        return self._duration

    @duration.setter
    def duration(self, _) -> None:
        path = list(bresenham(self.position, self.target.position))
        self._duration = len(path)
        self.position = path[0] if path else self.position

    def on_expire(self):
        self.target._take_damage(self.caster, self.damage_total, 'spirit', True, "Shade of Death")
        self.target.effects.add_stacks(ExposeWeakness, stacks=10, duration=5)
        self.target.effects.add_stacks(MagicVulnerability, stacks=10)
        self.target.effects.add_stacks(Frailty, stacks=10, duration=5)

class FrostResistance(Effect):
    def __init__(self) -> None:
        super().__init__(name="Frost Resistance", duration=20, category={
            'resist', 'ice', 'frost', 'frost resistance'}, take_bonus_damage_percent=-.10, symbol='🥶')


class FireResistance(Effect):
    def __init__(self) -> None:
        super().__init__(name="Fire Resistance", duration=20, category={
            'resist', 'fire', 'fire resistance'}, take_bonus_damage_percent=-.10, symbol='🥵')




#
# BUFFS

# ❤ (Cleric Heal)
# 💕 (Druid Heal)
# 💗 (Shaman Heal)
# 💖 (Used Cure)
# ⚡ (Used Stun?)
# 💙 (Has Heal over Time effect)
# 💚 (has DoT that can be cured)
# 🌀 (stunned)
# 🐌 (rooted) or 🌳
# 👀 (blind)
# ⚔ (attacked)
# 🦶 (moved)
# ❄ (has Frost exhaustion)
# 🔥 (has Fire exhaustion)
# ⚙ (other)

# 🥶 frost
# 🥵 fire
# 😵 stun
# 🤐 silence
# 🤢 sick/poison
# 🤪 confusion
# 🥴 weak
# 😑 blind

# Might (stacking) – Increase damage done by 5%
# Toughness (stacking) – Reduces damage received by 5%
# Next Attack – Increase damage done by X% on next attack (value on creation)
# Barrier – Absorb and store all damage for X turns (owned by caster,  object passed)
# Reflect – Reflect 50% of damage received back to target (can receive increasing modifiers)
# Parry – Completely avoid an attack
#!!-- Cure – Remove negative ailments from the target!!
# Shield Stance - Reduce damage by 75% for 3 turns. Prevents any action during this time. At the end receive 5 stacks of Toughness for each hit taken while Shield Stance was active.
#

from typing import Union
from .entities.characters import Party
from .entities.pawn import Pawn
from .entities.effects import Effect

from .debuffs import ExposeWeakness


class Parry(Effect):
    def __init__(self, user: Pawn, duration: Union[int,float] = float('inf')) -> None:
        super().__init__(name="Parry", duration=duration, take_bonus_damage_percent=-1.0,
                         category={'physical', 'buff', 'parry', 'defense', 'defensive', 'reflect'}, symbol='🖕')
        self.user = user
        self.user.effects.remove_name("might")
        self.user.effects.remove_name("expose weakness")

    def on_activate(self, *args, **kwargs) -> None:
        if isinstance((damager := kwargs.get('damager')), Pawn) and kwargs.get("total_damage", 0) > 0:
            self.duration = 0
            self.user.effects.add_stacks(Toughness, stacks=2, duration=3)
            damager.effects.add_stacks(ExposeWeakness, stacks=4, duration=3)


class Might(Effect):
    def __init__(self, duration: Union[int,float]) -> None:
        super().__init__(name="Might", duration=duration, deal_bonus_damage_percent=.05,
                         category={'physical', 'buff', 'might', 'strength', 'offensive'}, symbol='💪')


class Toughness(Effect):
    def __init__(self, duration: Union[int,float]) -> None:
        super().__init__(name="Toughness", duration=duration, take_bonus_damage_percent=-.05,
                         category={'physical', 'buff', 'tough', 'toughness', 'defense', 'defensive', 'modifier'}, symbol='✊')

class HoT(Effect):
    def __init__(self, name: str, duration:Union[int,float]=3, heal_amount:int=3) -> None:
        super().__init__(name=name, duration=duration, heal_over_time=heal_amount, category={'hot', 'buff', 'heal', 'heal over time'}, symbol='💕')

class CurativeNotes(Effect):
    def __init__(self, target: Pawn) -> None:
        self.target = target
        super().__init__(name="Curative Notes", duration=2,
                         category={'cure', 'buff'}, symbol='💊')

    def on_expire(self) -> None:
        self.target.effects.remove_all(
            *self.target.effects.get_damage_over_time_effects())


class NextAttack(Effect):
    def __init__(self, damage_percent: float) -> None:
        super().__init__(name="Next Attack", duration=float('inf'), deal_bonus_damage_percent=damage_percent,
                         category={'physical', 'buff', 'next attack', 'offense', 'offensive'}, symbol='💢')

    def on_activate(self, *args, **kwargs) -> None:
        self.duration = 0


class Barrier(Effect):
    def __init__(self, caster: Pawn, duration) -> None:
        super().__init__(name="Barrier", duration=duration, category={
            'physical', 'buff', 'barrier', 'shield', 'defense', 'defensive'}, symbol='🤍')
        self.caster = caster

    def on_activate(self, *args, **kwargs) -> None:
        self.caster.reports.setdefault('barrier', 0)
        self.caster.reports['barrier'] += kwargs.get("total_damage", 0)


class DivineProtection(Effect):
    def __init__(self, caster: Pawn, duration) -> None:
        super().__init__(name="Divine Protection", duration=duration, category={
            'physical', 'buff', 'barrier', 'shield', 'defense', 'defensive'}, symbol='🤍')
        self.caster = caster

    def on_activate(self, *args, **kwargs) -> None:
        self.caster.reports.setdefault('barrier', 0)
        self.caster.reports['barrier'] += kwargs.get("total_damage", 0)
        self.duration = 0


class Reflect(Effect):
    def __init__(self, duration: Union[int,float], extra: float = 0.) -> None:
        percent = .5 + extra
        super().__init__(name="Reflect", duration=duration, category={
            'physical', 'buff', 'reflect', 'defense', 'offense', 'defensive', 'offensive'}, symbol='♻', reflect_damage_percent=percent,)


class ShieldStance(Effect):
    def __init__(self, user: Pawn) -> None:
        super().__init__(name="Shield Stance", duration=3, category={
            'physical', 'buff', 'shield stance', 'shield', 'defense', 'defensive', 'damage_activate'}, symbol='🔰', 
            take_bonus_damage_percent=-.75)
        self.user = user
        self.user.reports['shield stance'] = 0
        
    def on_activate(self, *args, **kwargs) -> None:
        self.user.reports['shield stance'] += 1

    def on_expire(self) -> None:
        times = self.user.reports.get('shield stance', 0) * 5
        self.user.effects.add_stacks(Toughness, stacks=times, duration=3)


class Inspiration(Effect):
    def __init__(self, party: Party) -> None:
        super().__init__(name="Inspiration", duration=5, category={
            'physical', 'buff', 'inspiration', 'offense', 'offensive', 'damage_activate'}, symbol='🎶')
        self.party = party

    def on_activate(self, *args, **kwargs) -> None:
        for member in self.party:
            member._heal(kwargs.get("total_damage", 0))
            member.effects.add(Toughness(5))



import json
import copy
from typing import Any, Dict, List

try:
    import importlib.resources as resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as resources # type: ignore

from . import base_maps

from ..entities.board import Square, Board
from .location import Point


_TILE_MAP = {
    (0, 0, 0):{'symbol': '⬛', 'impassable': True, 'is_burning': False, 'is_lava': False, 'is_water': False, 'damage': 0}, # Black - Void/Impassable/Walls
    (255, 0, 0):{'symbol': '🟥', 'impassable': False, 'is_burning': False, 'is_lava': True, 'is_water': False, 'damage': 500}, # Red - Lava
    (0, 255, 0):{'symbol': '🟢', 'impassable': False, 'is_burning': False, 'is_lava': False, 'is_water': False, 'damage': 0}, # Green - Party Start
    (100, 100, 100):{'symbol':'🔴', 'impassable': False, 'is_burning': False, 'is_lava': False, 'is_water': False, 'damage': 0}, # Gray - Boss Start
    (0, 0, 255):{'symbol': '🟦', 'impassable': False, 'is_burning': False, 'is_lava': False, 'is_water': True, 'damage': 0}, # Blue - Water
    (255, 255, 0):{'symbol': '🟨', 'impassable': False, 'is_burning': True, 'is_lava': False, 'is_water': False, 'damage': 10}, # Yellow - Burning
    (255, 255, 255):{'symbol': '⬜', 'impassable': False, 'is_burning': False, 'is_lava': False, 'is_water': False, 'damage': 0}, # White - Floor
    (80, 50, 0):{'symbol': '🟫', 'impassable': True, 'is_burning': False, 'is_lava': False, 'is_water': False, 'damage': 0}, # Brown - Tree Trunk/Wooden object
    (0, 100, 0):{'symbol': '🟩', 'impassable': True, 'is_burning': False, 'is_lava': False, 'is_water': False, 'damage': 0}, # Green Square - Tree Leaves/Dense Foliage
}



def dict_to_square(d: Dict[str,Any]) -> Square:
    return Square(
        position=Point(*d['position']),
        symbol=d['symbol'],
        impassable=d['impassable'],
        is_burning=d['is_burning'],
        is_lava=d['is_lava'],
        is_water=d['is_water'],
        damage=d['damage'],
    )

def square_to_dict(s: Square) -> Dict[str,Any]:
    return {
        'position': [s.position.x, s.position.y],
        'symbol': s.symbol,
        'impassable': s.impassable,
        'is_burning': s.is_burning,
        'is_lava': s.is_lava,
        'is_water': s.is_water,
        'damage': s.damage,
    }

def json_to_board(json_str: str) -> Board:
    squares = json.loads(json_str)
    grid = []
    for _ in range(len(squares)):
        grid.append([])

    for row in squares:
        for square_dict in row:
            grid[square_dict['position'][1]].append(dict_to_square(square_dict))

    return Board(grid=grid)


def tiff_to_dict(read_loc: str, save: bool=False, save_loc: str='') -> List[Dict]:
    from PIL import Image

    im = Image.open(read_loc, 'r')
    width, height = im.size
    picture = im.load()
    js = []
    for y in range(height):
        row = []
        for x in range(width):
            square = copy.deepcopy(_TILE_MAP[picture[x,y]]) # type: ignore
            square.update({'position':(x,(height-y)-1)})
            row.append(square)

        js.append(row)
    
    if save and save_loc:
        with open(save_loc, 'w', encoding='utf8') as file:
            json.dump(js, file, ensure_ascii=False, indent=4)
    
    return js


def get_map(map_name: str) -> Board:
    """Returns a Board object for the given map name.

    :param map_name: the name of the map to load
    :return: a Board object for the given map
    """
    return json_to_board(resources.read_text(base_maps, map_name))